/*				FINDER.c
*
*  This is the top file for FINDER, except for Fmain.c where the
*  function main() lives. When FINDER is used as an external utility
*  by other programs, this file will be linked but Fmain.c will not.
*/


#define MAINFILE
#include "FINDER.h"

static int noclause;


#ifdef SOFTIE_SCOTT
#include "SOFTIE_SCOTT.h"
#endif

// Helicon gykA9y5A


/*
*  Found first tests the given clause in the current model.
*  If the clause is true in the model and add_if_true is set, then it
*  is added to the current theory if add_if_true is true.
*  If the clause is not true in the current model and seek_model is
*  set, then a model of the current theory plus the given clause is
*  sought. If one is found then it becomes the current model and the
*  given clause is added to the current theory.
*
*  The value returned is in every case the truth value of the clause
*  in the (possibly new) current model.
*
*  Note that the timer is reset on every call. This allows a time limit
*  to be set for each individual search.
* 
* Any output produced during the search will be directed by the
* verbosity and outfil settings in force when this function is called.
*/	

int found(const dynstring *s, int add_if_true, int seek_model)
{
  int i, oldgood, rc;
  functor f;

  if (no_semantics) return 0;

  //  signal(SIGALRM,short_timesup);  // May need to be there!

  if (seek_model) add_if_true = 1;         /* Sanity condition */

  /*
   * Now set things up so that s is the "latest clause" and the last
   * good model is ready for testing, and test the clause.
   */

  normal_input = false;
  finder_input(s);

  for (i=0; sort[i].sname; i++)
    sort[i].siz = last_sizes[current_theory][i];
  for (f=FN; f; f=f->nextfunct) {
    f->fcc = f->old_fcc[current_theory];
    set_fats(f);
  }
  set_table(FN);

  rc = test(last_model[current_theory],0);

  /*
   * If necessary, try to improve the model to accommodate the
   * new clause. This may require other actions such as re-testing
   * clauses (e.g. re-testing the set of support). Note that if the
   * "latest clause" is too long, no new model will be sought.
   */

  if (clause_is_long() && latest_clause.softweight) {
    seek_model = 0;
    add_if_true = 0;
  }

  if (seek_model && !rc) {
    tr_parameters.done = 0;
    tr_parameters.reps[NULL_BACKTRACKS].i_report = 0;
    tr_parameters.reps[SURJ_BACKTRACKS].i_report = 0;
    tr_parameters.reps[SOLUTIONS].i_report = 0;
    tr_parameters.reps[BAD_TESTS].i_report = 0;
    tr_parameters.reps[INVALUES].i_report = 0;
    oldgood = updates[current_theory];
    /*
     *    tr_parameters.max_penalty = theBadness + latest_clause.badness;
     */
    tr_parameters.max_penalty = 1;

    //   alarm(setting.maxtime);      Taken out for now - needs work!
    reset_timer();
    begin_time = timestamp();
    go(0,1);
    end_time = timestamp();
    //   alarm(0);

    if (updates[current_theory] > oldgood) {
      rc = found(s,0,0);
#ifdef SOFTIE_SCOT
      model_clauses = size[current_theory];
#endif
    }
  }

  if ((rc || latest_clause.softweight) && add_if_true)
    add_to_theory();
  normal_input = true;

  return (rc);
}



void init_init_finder(void)
{
  no_semantics = false;
  job_only = false;
  noclause = false;
#ifdef SOFTIE_SCOT
  init_buckets();
#endif
  updates[0] = 0;
}


/*
*	Init_finder must be called before much else happens if FINDER
*	is to be used as a supertester. It reads the input file and
*	returns a boolean (integer) value recording whether it found
*	any model of the input theory with badness no worse than the
*	level to be "tolerated". If the return value is false then
*	there will be no "current theory", or the current theory will
*	be intolerably bad, and adding clauses may be a bad idea.
*/

boolean init_finder(char s[], int tolerated)
{
  FILE *if_fil;

  if (s) {
    if ( !(if_fil = fopen(s,"r")) )
      inerror("Unable to open file",s,0);
    return _init_finder(if_fil,0,0,tolerated);
  }
  return false;
}

boolean initFinder(ifile i, ifile o, int tolerated)
{
  return _init_finder(0,i,o,tolerated);
}

boolean _init_finder(FILE *if_fil, ifile ifili, ifile ifilo, int tolerated)
{
  int i;

#ifdef SOFTIE_SCOTT
  if (Flags[NO_SEMANTICS].val) {
    no_semantics = true;
    return 0;
  }
  CLOCK_START(FINDER_TIME);
#endif
  
  //  signal( SIGALRM, short_timesup );
  initialize();
  if (noclause)
    setting.parse_only = true;
  if (job_only)
    setting.clausify_only = true;

  if (if_fil) {
    read_job(if_fil,0);
    fclose(if_fil);
  }
  else if (ifili)
    read_job(0,ifili);
  if (ifilo)
    setting.outi = ifilo;
  else if (!setting.outf)
    setting.outf = stdout;

  tr_parameters.done = 0;
  for (i=0; i<MAXISET_ARRAY; i++) updates[i] = 0;
  single_model = true;
  if ( setting.pm.verbosity[pjob] || job_only ) {
    print_the_job();
    if (job_only)
      return 1;
  }

  current_theory = 0;
  reset_timer();
  begin_time = timestamp();
#ifdef SOFTIE_SCOTT
  set_maxtime(setting.maxtime * 3);
#else
  set_maxtime(setting.maxtime);
#endif      
  go(0,1);
  end_time = timestamp();

#ifdef SOFTIE_SCOTT
  CLOCK_STOP(FINDER_TIME);
#endif
  if (setting.pm.verbosity[pstats])
    printstats();
  return (initial_model_found() && last_badness <= tolerated);
}


/*
 * Cut-down version of the above, to get the job into standard format
 * in a new ifile.
 */

ifile transfer_job(char *source)
{
  FILE *if_fil;
  ifile of_ifil;

  initialize();
  if (!(if_fil = fopen(source,"r")))
    inerror("Unable to open file",source,0);
  read_job(if_fil,0);
  fclose(if_fil);
  of_ifil = iopen();
  xprint_the_job(0,of_ifil);
  return of_ifil;
}




/*
*	FINDER can be used to return models falsifying a badguy.
*/

void refute_current(void)
{
  cc_badguy = true;
}




/*
*	It is sometimes useful for a calling program to know how many 
*	clauses there are in a theory.
*/

int theory_size(int x)
{
  int i = 0;
  clause c;

  for (c=TH[x]; c; c=c->nextclause) i++;
  return i;
}


void short_timesup(int x)
{
  job_done(&tr_parameters,TIMEOUT);
}


boolean initial_model_found(void)
{
  return (updates[0] > 0);
}


void set_no_semantics(void)
{
  no_semantics = true;
}


int no_semantics_set(void)
{
  if (no_semantics) return 1;
  return 0;
}


void set_job_only(void)
{
  job_only = true;
}


void unset_job_only(void)
{
  job_only = false;
  noclause = false;
}


int job_only_set(void)
{
  if (job_only) return 1;
  return 0;
}


void set_noclause(void)
{
  noclause = true;
  job_only = true;
}


void unset_noclause(void)
{
  noclause = false;
}


int noclause_set(void)
{
  if (noclause) return 1;
  return 0;
}


/*
 * Free any memory allocated and not otherwise freed.
 * This seems to be just wffstacks (varlists, testlists, prem and conc lists)
 */

void FINDER_cleanup(void)
{
  int x;
  clause c;

  clean_up_cc();
  for (x=0; x<MAXISET_ARRAY; x++)
    for (c=TH[x]; c; c=c->nextclause)
      free_clause_parts(c);
}




void print_model_anyway(void)
// Not currently called from anywhere. Maybe used in SCOTT.
{
  int i;
  functor f;

  for ( i = 0; sort[i].sname; i++ )
    sort[i].siz = last_sizes[current_theory][i];
  for ( f = FN; f; f = f->nextfunct ) {
    f->fcc = f->old_fcc[current_theory];
    set_fats(f);
  }
  vect_into_model(last_model[current_theory]);

  print_the_model();
}


void stop_code(char *dest)
// Print in dest the reason why the last search stopped.
// Assumes that the trin is still active and that dest is long enough.
// May need to clear trailing space or line feed from dest.
{
  if (tr_parameters.done) {
    strcpy(dest,tr_parameters.reps[STOP_CODE].s_report);
    while (isspace(dest[strlen(dest)-1]))
      dest[strlen(dest)-1] = 0;
  }
  else
    strcpy(dest,"exhaustion");
}


int finder_solutions(void)
{
  return tr_parameters.reps[SOLUTIONS].i_report;
}
