/*------------------------------------------------------------------------*/
/*
 *                   Type definitions for FINDER 
 */
/*------------------------------------------------------------------------*/


typedef enum   { None, Func, Quant, Var, Const } syntype;
typedef enum   { null, brief, full } prints;
typedef enum   { nul, srt, fcn, fcn0, vcb, cls, prm, inc } statement_type;
typedef enum   { math, comp } spec_type;

typedef enum {
  pmodels,
  pjob,
  pstats,
  pspace,
  pmodes
} print_mode;

/*--------------------- a list of cannonical names ------------------------*/
typedef struct s_name_list {
  char *name;
  struct s_name_list *nextname;
} NAME_LIST, *name_list;


/*----------------------------- a sort ------------------------------------*/
typedef struct {
  char *sname;          /* The identifier for this sort */
  name_list canonical;  /* List of names, for enumerated sort */
  int sizmin;           /* Least number of things of this sort */
  int sizmax;           /* Greatest ditto */
  int siz;              /* Current number of such things */
  int maxvalue;         /* Always set to siz-1 */
  int *sum;             /* Addition is a built-in function (array) */
  int *minus;           /* Subtraction is a built-in function (array) */
  int *pred;            /* Predecessor is a built-in function (array) */
  int *succ;            /* Sucessor is a built-in function (array) */
  int *Ex;              /* "Exists" is a predicate after all! (array) */
  int *len;             /* Name lengths of the elements (array) */
  boolean piped;        /* Sort "siz" comes via the pipe */
  boolean is_val;       /* Sort is used for a decision variable */
  set sizmask;          /* Bit mask for range 0..siz */
} KIND, *kind;

/*----------------------- a model of a functor ----------------------------*/
typedef struct s_functor {
  char *fname;                  /* The identifier for this symbol */
  char *pretty;                 /* Alternative I/O string */
  syntype category;             /* Syntactic type (func, pred, etc) */
  int scope;                    /* For parsing purposes */
  int adicity;                  /* Number of argument places */
  int *table;                   /* Start of matrix for this function */
  int fcc;                      /* Index: start of portion of vector */
  int old_fcc[MAXISET_ARRAY];   /* Was fcc on last good model */
  int fats[ARGMAX];             /* For use in matrix lookup */
  int binder;                   /* Binding quantifier, if it's a variable */
  boolean total;                /* Function may not have null value */
  boolean injective;            /* Function is one-one */
  boolean leftin;               /* Each row is one-one */
  boolean rightin;              /* Each column is one-one */
  boolean surjective;           /* Function is onto */
  boolean leftsur;              /* Each row is onto */
  boolean rightsur;             /* Each column is onto */
  boolean surj_set;             /* Skolem mate(s) for surjection defined */
  boolean comm;                 /* Function is binary and commutative */
  boolean piped;                /* Function values come via the pipe */
  boolean nop;                  /* No decrease in change-order priority */
  boolean cut_it;               /* Prohibit backtracking on function */
  boolean labels;               /* Label the printout with names */
  boolean cgv;                  /* May be reported for visualisation */
  boolean configurable;         /* Report possible values, not solutions */
  boolean constant_table;       /* Memory for table should not be freed */
  boolean declared;             /* Symbol declared in vocabulary section */
  kind arg_type[ARGMAX];        /* The types of the argument(s) */
  kind val_type;                /* The type of this function's value */
  prints    pmode;              /* What kind of print display */
  struct s_functor *nextfunct;  /* Linking the list of "real" functors */
} FUNCTOR, *functor;
   

/*----------------------- a well formed formula ----------------------------*/
typedef struct s_wff {
  functor sym;                /* Main predicate or function symbol */
  struct s_wff *sb[ARGMAX];   /* Offsets of the subformulae */
  int val;                    /* Value currently assigned */
  int *matrix;                /* Start of the relevant value table */
  int *sbval[ARGMAX];         /* Pointers to values of subformulae */
  int wcc;                    /* Offset in info vector for t-testing */
  struct s_wff *nextwff;      /* Linking the list of testables */
} WFF, *wff;


/*-----------------a stack of well formed formulae -------------------------*/
typedef struct s_wff_stack {
  wff w;
  struct s_wff_stack *nxt;
} WFF_STACK, *wff_stack;


/*-----------------a stack of function symbols -------------------------*/
typedef struct s_funct_stack {
  functor f;
  struct s_funct_stack *nxt;
} F_STACK, *f_stack;


/*----------------- clause as a pair of wff lists: prem and conc -----------*/
typedef struct s_clause {
  int cid;                       /* clause id */
  int pid;                       /* parent clause id */
  wff_stack prem;                /* negative literals */
  wff_stack conc;                /* positive literals */
  wff_stack varlist;             /* variables in the clause */
  wff_stack testlist;            /* subformuale in test order */
  int weight;                    /* number of function symbols */
  int varcount;                  /* number of variables */
  int tuples;                    /* upper bound on constraints (or 10^6) */
  int softweight;                /* cost of each violation (0 if hard) */
  int badness;                   /* sum of current violation costs */
  boolean preprocessed;          /* already treated as small clause */
  boolean tautology;             /* known to be useless */
  struct s_clause *nextclause;
  struct s_clause *lastclause;
} CLAUSE, *clause;


/*-------------------------- semi-parsed formulas --------------------------*/
typedef struct s_psw {
  functor main;
  wff actual;
  struct s_psw *leftsub, *rightsub;
  boolean m;
} PSW, *psw;


/*---------------------a meaningful word in the language -------------------*/
typedef struct {
  char word[LEXLENGTH];
  boolean iskind;            /* flag if word represents a sort */ 
  int offset;                /* its position in the relevant array */
  int alias;                 /* lex item to which it converts */
} LEXITEM;


/*------------------list of variable bindings-------------------------------*/

typedef struct s_bdl {
  char v[LEXLENGTH];         /* string variable representatiom */
  int q;                     /* id of quantifier binding it */
  struct s_bdl *next;
} BINDING_LIST, *binding_list;


/*----------------------------PARAMETERS------------------------------------*/

typedef struct {
  int max_models;            /* Limit on number of solutions */
  int maxref;                /* Above which "notop" deletions happen */
  int test_maxref;           /* Same for test of long clauses */
  int max_backtracks;        /* Break condition used in SCOTT */
  int max_invalues;          /* Break condition used in SCOTT */
  int max_flips;             /* Break condition for local search */
  int hardweight;            /* For local search */
  int pretest_limit;         /* Above which lazy constraint generation */
  int batches;               /* For determining when to cut and guess */
  int extra_batches;         /* Something or other */
  int target_bound;          /* Only count solutions better than this */
  int dump_rate;             /* Frequency of trace dumping (default 0) */
  int max_cellvals;          /* Maximum job size in c-v pairs (if set) */
  logical maxtime;           /* Break condition - elapsed time limit */
  boolean parse_only;        /* Check syntax but don't clausify or solve */
  boolean clausify_only;     /* Parse and clausify but don't solve */
  boolean search_is_local;   /* Local search only: no backtracking */
  boolean Nosecs;            /* No dependency-direction */
  boolean invert_change;     /* Variable values increase in test */
  boolean vars_over_wt;      /* Variable count beats weight in clause order */
  boolean topsoff;           /* Old feature - to be deleted */
  boolean report_breaks;     /* Print reason why search stopped */
  boolean print_headers;     /* Write labels "Model N" etc */
  boolean noisy;             /* Beep when a model found */
  boolean sortcut;           /* Don't increase domain size if models found */
  boolean timestamps;        /* Print the time on each model found */
  boolean random;            /* Values to be inserted randomly */
  boolean profile;           /* Print values for visualisation */
  boolean dumpdepth;         /* Diagnostics: record backtrack depths */
  boolean awnty_enabled;     /* Print data for tree size estimator */
  boolean cgv_enabled;       /* Print data for search visualisation */
  boolean debug_enabled;     /* Print data for debugger */
  boolean configmode;        /* Run in "product configuration" mode */
  boolean configheuristic;   /* In configuration" mode, do sensible things */
  boolean single_literal;    /* FOF input only - no multi-literal clauses */
  boolean drilldown;         /* Give detailed error messages where possible */
  boolean print_unsat;       /* Print clauses unsatisfied by approx model */
  spec_type function_spec_mode; /* Syntax for declarations */
  FILE *outf;                /* Where output goes (an external file) */
  ifile outi;                /* Where output goes (an internal file) */

  struct printmode {
    prints verbosity[pmodes]; /* Print job, models, stats to order */
    boolean pretty;           /* Human-readable output */
    boolean incremental;      /* Print suboptimal solutions as they improve */
    boolean number_only;      /* Print only the number of solutions */
  } pm;
} PARAMETERS;

/*=========================== END =========================================*/

