
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
// #include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <unistd.h>

#include "debugger.h"
#include "sets.h"
#ifndef IFILES
#include "ifile.h"
#endif

/*
 *   Where the error messages go to (usually stderr, or stdout for web use)
 */

#ifdef WORLDWIDE
#define WARNING_STREAM stdout
#else
#define WARNING_STREAM stderr
#endif

/*
* Constant definitions
*/

#define F_MAXINT 1000000000
#define REF_MAX 10000
#define REF_DEFAULT (REF_MAX-1)
#define EXTRA_DEFAULT 500
#define MAX_REF_SIZE 8
#define STAK_MAX (REF_MAX*MAX_REF_SIZE)
#define MAX_PENALTY F_MAXINT
#define LAST_FLIP_INIT 100
#ifdef WORLDWIDE
#define CELLVAL_MAX 100001
#else
#define CELLVAL_MAX 0             // Number of cell-value pairs (if set)
#endif

/* This is mandatory (set it to the right value) if CLOCKS_PER_SEC is
   not found
*/
#define LOCAL_CLK_TCK 1000

/*
 * WSZ (word size) is the number of bits in a "logical" variable.
 */

#define WSZ 32

#define SURJ_MAX 256
#define RRA 50
#define RRB 30

#define NO_REPORT ((unsigned)1)
#define INT_REPORT ((unsigned)2)
#define FLOAT_REPORT ((unsigned)4)
#define DO_REPORT ((unsigned)8)

#define logical unsigned int

#define FOR(x,y) for( y = x; y; y = y->p )
#define FORR(x,y) FOR( x->cellvals, y )

/*
* Type definitions
*/

#ifndef BOOL
#define BOOL
#define boolean char
#define false ((boolean)0)
#define true ((boolean)1)
#endif

typedef enum {
  NOCUT,
  FOUND,
  RECORDED,
  COMPLETED
} cutcode;

typedef enum {
  PERFECT,
  NULL_SPACE,
  BIG_SPACE,
  REF_LIST,
  BUST,
  IBUST,
  NULL_REF,
  SURJES,
  SKIP
} exit_code;

typedef enum {
  EXHAUSTION,
  TIMEOUT,
  ENOUGH_SOLUTIONS,
  ASSIGNMENT_OK,
  INSERTIONS,
  BACKTRACKS,
  INTERRUPT
} termination_code;

/*
 * report codes defined in debugger.h
 */

typedef struct IL {
  boolean positive;             /* Represents positive literal */
  int v, w;                     /* A couple of integers */
  struct IL *p;                 /* in a singly linked list */
} INTPAIR, *intpair;


typedef struct reftype {
  int cardinality;             /* Just what is says */
  boolean notop;               /* Useless extension omitted */
  intpair cellvals;            /* The cell-valuess involved */
  int softweight;              /* Cost if violated if a soft constraint */
  boolean violated;            /* All variables assigned and clause false */
  int last_flip;               /* When last chosen by local search */
  struct reftype *nxtr;        /* Just for stack management */
} REF, *ref;


typedef struct rf_list {        /* A list of refutations  */
  ref rf;
  struct rf_list *nextref;
  struct rf_list *lastref;      /* Added for local search */
} REFLIST, *reflist;


typedef struct {
  int blocks;                 /* How many times blocked */
  ref back;                   /* For calculating secondaries */
  reflist cvrefs;             /* The refs involving this c-v */
  reflist cvrefs_eol;         /* End of cvrefs - for local search */
  reflist posrefs;            /* Positives involving this c-v */
  reflist posrefs_eol;        /* End of posrefs - for local search */
  int vpos;                   /* Offset of its cell */
  int penalty;                /* Soft cost of inserting this value */
  int last_flip;              /* When last chosen by local search */
  boolean known_possible;     /* Occurs in at least one found solution */
} CELLVALUE, *cellvalue;


typedef struct {
  int poss;                   /* How many possible values */
  int csiz;                   /* Cardinality of {0, ... max value} */
  int vpos;                   /* Vector position of this cell */
  reflist inject;             /* Index to injective function */
  CELLVALUE *cvals;           /* The cell-value structures (array) */
  boolean used;               /* Involved in new refutation */
  boolean has_value;          /* This cell has a value */
  boolean fixed;              /* Has only one possible value */
  boolean allknown;           /* Every value known to be possible */
  int times_picked;           /* How often this cell split upon or flipped */
  int last_flip;              /* When last chosen by local search */
  int least_penalty;          /* Penalty of the best available value */
} CELL, *cell;


typedef struct {
  set valset;                 /* Values which must occur somewhere */
  int *last_seen;             /* Which cell witnesses value (array 0..siz) */
  intpair cellset;            /* Where the values must occur */
} SURJ;


typedef struct {
  logical report_type;
  char s_report[80];
  int i_report;
  float f_report;
  void (*do_report)();
} REPORT;


typedef struct cng {
  int forkpoint;             /* Cell where division takes place */
  int cutval;                /* Value for that cell */
  int *icomvec;              /* Current values of all (significant) cells */
  int *inorder;              /* Order in which they should be placed */
  set *possvals;             /* Domains including blocked values */
  set *valsleft;             /* Values waiting to be tried */
  boolean active;            /* Cut is currently being applied */
  struct cng *lastone;
} CUTT, *cutt;

#define NEWSLENGTH 1024

typedef struct { REF news[NEWSLENGTH]; }  RBATCH,  *rbatch;
typedef struct { INTPAIR news[NEWSLENGTH]; }  ILBATCH, *ilbatch;
typedef struct { REFLIST news[NEWSLENGTH]; }  RLBATCH, *rlbatch;


typedef struct {
  boolean (*Test)();            /* The principal test function  */
  boolean (*Onerefs)();         /* Removal of impossible values */
  boolean (*Smallrefs)();       /* Special treatment for smalls */
  void (*Sizes)();              /* Get the domain sizes */
  void (*Printdomains)();       /* Output for config mode */
  int vlength;                  /* Local vector length */
  int maxval;                   /* Local max domain size */
  int batches;                  /* Local stack size / 1000 */
  int extra_batches;            /* Batches after pretest */
  int maxref;                   /* Maximum refutation size */
  int max_cellvals;             /* Maximum cell-value pairs (if set) */
  unsigned maxtime;             /* Time limit (only used in config mode */
  int maxbak;                   /* Maximum number of backtracks */
  int maxinvals;                /* Maximum number of value insertions */
  int maxflips;                 /* For local search */
  int verb;                     /* Statistics verbosity setting */
  int ref_record[RRA][RRB];
  int *branches;                /* For stats (array 0..siz) */
  int max_penalty;              /* Maximum badness allowable */
  int hardweight;               /* Badness of hard constraints (local search) */
  int dump_rate;                /* How often to dump a trace */
  int tracepoints;              /* How often trace dump point reached */
  boolean search_is_local;      /* Local search, not backtracking */
  boolean piping;               /* Some functions come from the pipe */
  boolean Nosecs;               /* Suppress secondary refs */
  boolean topsoff;              /* Remove redundant ref tops */
  boolean report_breaks;        /* Print reason for end search */
  boolean print_headers;        /* Print labels "Model N" etc */
  boolean dumpdepth;            /* Print backtracking depths */
  boolean awnty_enabled;        /* Print search tree estimate data */
  boolean debug_enabled;        /* Print search tree estimate data */
  boolean cgv_enabled;          /* Print data for search visualisation */
  boolean *forbidden;           /* No backtracking (array 0..vlength) */
  boolean *configurable;        /* In config mode, "interesting" cell */
  boolean random;               /* Insert values in random order */
  boolean done;                 /* End-of-search status raised */
  boolean profile;              /* Print info for visualisation  */
  boolean configmode;           /* Test for really possible values */
  boolean configheuristic;      /* In config mode, act sensibly */
  FILE *outf;                   /* Output file (external) */
  ifile outi;                   /* Output file (internal) */
  REPORT  reps[MAX_REPORT+1];   /* Search statistics */
  char tcode[INTERRUPT+1][100]; /* Reasons for termination */
} TRIN, *trin;



typedef struct {
  trin job;
  cell vector;                  /* The decision variables (array) */
  set *scomvec;                 /* Set communication vector (array) */
                                /* - during search, remaining poss values */
  int *icomvec;                 /* Integer communication vector (array) */
                                /* - during search, current values */
  int *inorder;                 /* When values were inserted (array) */
  int *coinorder;               /* Converse of inorder (array) */
  int *priority;                /* Level of function declarations (array) */
  int *best;                    /* Lowest penalty possible for cell (array) */
  int best_penalty;             /* Lower bound on sum of penalties */
  int total_penalty;            /* Actual penalty for inserted (array) values */
  int violations;               /* Violated constraints, for local search */
  int sdc_violations;           /* Soft domain constraint violations */
  int flips;                    /* Number of random steps, for local search */
  int *awnty_past;              /* Number of subtrees searched (array) */
  int *awnty_future;            /* Number (possibly) to be searched (array) */
  int *awnty_sofar;             /* Cumulative subtree size */
  int this_cell;                /* Ordinal of current insertion */
  int susguy;                   /* Wiped out cell (for finding nogood) */
  int firstchange;              /* Most significant change since last model */
  int forkpoint;                /* Most significant really branching node */
  int maxval;                   /* Max domain size for any cell */
  int stakmax;                  /* Max batches * max ref size */
  int unfixt;                   /* Highest numbered cell with no value */
  int local_memory;             /* Current size of constraint database */
  int max_batches;              /* Bounds number of batches of constraints */
  int configtestcell;           /* In config mode, cell being tested */
  int configtestval;            /* In config mode, value being tested */
  boolean configstart;          /* Putting in first value in config mode */
  boolean novect;               /* "No vector": no solution exists */
  boolean ts;                   /* "Totally suspended": wipeout detecte */
  boolean cut_recorded;         /* Search space division has been processed */
  rbatch stak[REF_MAX];                  /* Reservoir topup chunks */
  ilbatch il_stak[STAK_MAX];             /* Reservoir topup chunks */
  rlbatch rl_stak[STAK_MAX + REF_MAX];   /* Reservoir topup chunks */
  ref stak_ptr;                          /* Reservoir */
  intpair il_free;                       /* Reservoir */
  reflist rl_free;                       /* Reservoir */
  reflist duds;                 /* Expired constraints (marked for deletion) */
  reflist viol_list;            /* Violated constraints: local search only */
  ref newref;                   /* Next constraint to be incorporated */
  SURJ surjection[SURJ_MAX];    /* Sets of cells in surjective functions */
  cutt cutptr;                  /* List of search space divisions */
  cutcode cut_flag;             /* Status of search space division */
  boolean absolute;             /* No real branching has happened yet */
} TRS, *trs;



/*
*  Global variables
*/

#ifdef TRFILE
#define TR_GLOB
#else
#define TR_GLOB extern
#endif

//---------------------- Annoying but needed ---------------------------//

TR_GLOB int theClockTick;

TR_GLOB trs debugger;
TR_GLOB int begin_time, end_time;

TR_GLOB int *debug_int;

#ifdef DEBUG
typedef struct db_ref {
  int where;
  REF r;
  int refid;
  struct db_ref *next;
  struct db_ref *last;
} DBREF, *dbref;
#define dbringmax 30
TR_GLOB DBREF REFRING[dbringmax];
TR_GLOB dbref refring;
void init_init_refring(void);
void init_refring(trs T);
void ref2ring(trs T, int w, ref rf);
void copy_ref(trs T, ref dest, ref source);
intpair rec_copy_ils(trs T, intpair source);
void dump_refring(void);
void rec_dump_refring(dbref current, dbref final);
#endif

TR_GLOB int *tr_last_vector;
TR_GLOB int tr_last_vlen;
TR_GLOB int last_badness;

/*
*  Function prototypes
*/


void transref(trin job);
int found_solutions(trs T);
void init_config(trs T, set **possibles, set **impossibles);
void initial_domains(trs T, set possibles[], set impossibles[]);
void set_domains(trs T, set possibles[], set impossibles[]);
void config_search(trs T, set possibles[], set impossibles[], int sol_found);
boolean find_unknown(trs T, boolean tabu[]);
void free_trs(trs T);
void free_config(set possibles[], set impossibles[], trs T);
void init_trin(trin tr);
void clean_up(trs T);
void termination_initial(trs T);
void report_initial(trs T);
trs tr_initial(trin job);
int search(trs T, int dpth, int bdpth);
int remaining(cell c, int wyord[], int bound);
int pick_cell(trs T);
int set_fork( trs T, int rc);
void test_possible_solution(trs T);
int make_search_tree_node(trs T, int dpth, int bdpth, int ex);
int get_possible_values(trs T, int ex, int wyord[]);
boolean set_wy(trs T, int wyord[], int jay, int ex, int bdpth,
	       int *kay, int *wy);
int recursive_search_call(trs T, boolean nochoice, int dpth, int bdpth);
void do_dumps(trs T, int dpth, int bdpth, int choice, int rc);
void randomise(int arr[], int bound);
void greedy_order(int arr[], int bound, int pos, cell c, trs T);
void unknowns_first(int wyord[], int length, cellvalue cv);
void delete_cut(trs T);
reflist get_rl(trs T);
void put_rl(trs T, reflist dead);
intpair get_il(trs T, boolean b, int x, int y, intpair ip);
void put_il(trs T, intpair dead);
int batches_used(trs T);
ref get_ref(trs T);
void put_ref(trs T, ref dead, boolean flag);
void rec_put_il(trs T, intpair ip, ref r, boolean flag);
void more_refs(trs T);
void more_rls(trs T);
void more_ils(trs T);
boolean invalue(trs T, cell c, int v);
void transfer_ref (trs T, ref rf);
void blockval(trs T, cell c, int v, ref rrf, report_code RC);
boolean surjok(trs T);
void find_min_penalty(cell c, trs T);
void set_total_penalty(trs T);
void outvalue(trs T, cell c);
void untransfer_ref (trs T, cell c, ref rf);
void unblockval(trs T, cell c, int v, ref rrf);
void mark_delete(trs T, ref r);
void delete_duds(trs T);
void setref(trs T);
void reduceref(trs T);
void addref(int x, trs T);
void totsus(trs T);
void no_topless_singles(trs T);
void make_coherent(trs T);
void no_ref(trs T);
void new_small_ref(int cells[], int vals[], boolean valency[],
		   int softweight, trs T);
void remove_value(trs T, cell c, int x);
intpair get_all_ils(int cells[], int vals[], boolean valency[], int x, trs T);
boolean subsumed(ref the_ref, trs T);
boolean subsumes(ref little_ref, ref big_ref);
void back_sub(trs T, ref the_ref);
boolean identical_ref( ref the_ref, trs T );
boolean is_the_same( ref the_new_ref, ref the_old_ref );
void Ref(int x, trs T);
void record_cut(trs T);
void record_part_cut(set s, trs T);
void complete_cut(trs T);
boolean prepared(trs T);
void setinx(trs T);
boolean pre_test(trs T);
boolean absfixt(int x);
void surjective(boolean vec[], set vset, trs T);
void injective(boolean vec[], int x, trs T);
void co_priority(int x, int y, trs T);
void ref_size_report();
void report_branches(trin tr);
void stopped_by( trin tr );
void dump_refs(int rows, int columns);
void job_done(trin t, termination_code tc);
void skipout(char s[], int exitcode);
void add_penalty(int vp, int val, int x, trs T);
void subtract_penalty(int vp, int val, int x, trs T);
void set_best(trs T);
boolean remove_singles(trs T);
boolean apply_unit_propagation(trs T, int x, int y, ref r);
void dump_depth(int d);
void dump_awnty(int past[], int future[], int sofar[], int depth, boolean term);
void set_last_vector(int thisvector[], int vlength);

/*
 * These in fprepare.c
 */

void print_un_vec_value(int x, int val);   // No longer exists
void report_constraint(ref thisref);
void report_literal(intpair lp);
void report_constraint_list(reflist theserefs);

/*
 * These in foutput.c
 */

void xprint(FILE *f, ifile i, char *s);
void dump_trace(trin tr, boolean final);
void print_constraint_graph(trs T);
void check_new_ref(int x, ref rf, trs T);
void cgv_state_report(trs T, cell c, int val);
void dump_state(set domains, trs T);
void print_cgv_set(set domain, int maxval);
void print_val_of_cell(FILE *fil, ifile ifil, int x, int v);
void print_un_vec_pos(FILE *fil, ifile ifil, int x);

/*
 * These in localsearch.c
 */

void localsearch(trs T);
void set_parameters(int *alt, int *reps);
void set_initial_assignment(trs T);
boolean still_searching(trs T);
void random_step(trs T, int reps, int age);
ref choose_random_constraint(trs T);
ref choose_oldest_constraint(trs T);
int choose_random_variable(trs T, ref r);
int choose_oldest_variable(trs T, ref r);
int choose_random_value(cell c, int current_value);
int choose_oldest_value(trs T, cell c, int current_value);
void calculate_penalty(trs T);
void local_setvalue(trs T, cell c, int v);
void add_positive_violations(trs T);
boolean starts_here(ref r, int h);
boolean is_violated(int vec[], ref r);
void set_eols(trs T);
reflist alldiff_violations(trs T, ref r, int ch, int v1, int v2, reflist supp);
reflist initial_alldiff_violations(trs T, ref r);
void merge_sort(int c[], int v[], int length);
int check_viol_list(trs T, ref r);
void sanity_check_viols(trs T);

/*
 * In debugger.c
 */

void EVENT(event_id e, trin tr);

/*
 * In diagnose
 */

int hardweight(void);

/*
 * In timing.c
 */

void reset_timer(void);
void set_maxtime(int s);
int time_limit(void);
int timeout(trs T);
int timestamp();
int timestat(trin Job, int old_t, report_code rc);


/*

void freeset(set s);
void really_free_sets(void);

*/
