/*
 * This contains the high level dialogue: procedures for setting up
 * nodes of the dialogue tree, asking for and interpreting user
 * choices from the menu of options, maintaining a record of the
 * dialogue and calling the appropriate lower level actions.
 */

#include "diagnose.h"
void initial_initial(void);

static int next_node_id = 1;       // Just generating unique id numbers


dialogue get_dialogue_node(int co)
// Allocate memory for a node of the dialogue, initialise and return it
{
  dialogue D = (dialogue) malloc(sizeof(DIALOGUE));

  D->absolute_node_id = next_node_id++;
  D->relative_node_id = 0;
  D->e = 0;
  D->nclauses = 0;
  D->theClauses = 0;
  D->nsubs = 0;
  D->sub = 0;
  D->theCore = 0;
  D->theModelFile = 0;
  D->seed = defaultSeed;
  D->time = defaultTimelimit;
  D->flip = defaultFliplimit;
  D->parent = 0;
  D->core_only = co;
  return D;
}


dialogue set_up_diagnose(int argc, char *argv[])
// Create the root node of the dialogue tree.
// Parameters: argc should be 2 or 3 and argv[argc-1] a .fin filename
{
  dialogue D;
  char output_mode[SLEN];
  char *infilname = 0;
  ifile tempi = iopen();
  boolean dummy1, dummy2;

  initial_initial();
  get_options(argc,argv,&dummy1,0,output_mode,&dummy2);
  infilname = argv[argc-1];

  // TODO: give D the input formulae as well as the constraints
  // TODO: if called with -o take softweights from the input
  INSERT_FACTOR = insert_factor_default;
  set_up_event_names();
  if (!strcmp(output_mode,"core"))
    D = get_dialogue_node(1);
  else
  if (!strcmp(output_mode,"optimal"))
    D = get_dialogue_node(2);
  else
    D = get_dialogue_node(0);

  tempi = transfer_job(infilname);
  problem_basis(tempi);
  D->theClauses = problem_clauses(tempi,&(D->nclauses));
  return D;
}


dialogue do_new_dialogue(dialogue p, diagnose_event cause)
// Make a node of the dialogue tree. Parameters are
//   pointer to the parent node
//   the event causing this node
// Note that a "no_op" event must be added to the parent's event
// list so that this portion of the tree can be correctly located
// in the sequence of events constituting the session.
{
  dialogue D;
  int x;
  event_list Ep;

  if (!p)
    abort_diagnose("Dialogue node has no parent",0);
  D = get_dialogue_node(0);
  D->parent = p;
  p->nsubs++;
  D->relative_node_id = p->nsubs;
  p->sub = (dialogue*) realloc(p->sub,(p->nsubs)*sizeof(dialogue));
  p->sub[p->nsubs-1] = D;
  Ep = (event_list) malloc(sizeof(EVENT_LIST));
  Ep->e = no_op;
  Ep->container = p->absolute_node_id;
  Ep->iparams = (int*) malloc(sizeof(int));
  Ep->iparams[0] = p->nsubs-1;
  Ep->sparams[0] = Ep->sparams[1] = 0;
  Ep->mresult = 0;
  Ep->cresult = 0;
  Ep->sresult = 0;
  Ep->next = 0;
  if (!p->e) {
    p->e = Ep;
    Ep->last = 0;
    Ep->event_id = 1;
  }
  else {
    for (Ep->last = p->e; Ep->last->next; Ep->last = Ep->last->next) ;
    Ep->last->next = Ep;
    Ep->event_id = Ep->last->event_id + 1;
  }
  D->nclauses = p->nclauses;
  D->theClauses = (dclause) malloc(D->nclauses * sizeof(DCLAUSE));
  for (x=0; x<=D->nclauses; x++)
    copy_dclause(p->theClauses+x,D->theClauses+x);
  process_event(D,cause);
  do_dialogue(D);
  return D;
}


void do_dialogue(dialogue D)
// Repeatedly take user choices of event and perform the actions.
// When an action changes the problem, start a new node.
{
  diagnose_event e;

  active_node = D;
  switch (D->core_only) {
  case 0:
    do {
      e = next_diagnose_event();
      switch(direction(e)) {
      case stay:
	process_event(D,e);
	break;
      case forward:
	do_new_dialogue(D,e);
	break;
      case back:
	switch(e) {
	case revert:
	  new_active_node(D);
	  break;
	case halt:
	  active_node = 0;
	  break;
	default:
	  break;
	}
      }
    }
    while (active_node == D);
    break;
  case 1:
    process_event(D,core);
    active_node = 0;
    break;
  case 2:
    process_event(D,optcore);
    active_node = 0;
    break;
  default:
    abort_diagnose("Unrecognised single command",0);
  }
}


void new_active_node(dialogue D)
// Action when user has chosen to return to an earlier node. The chosen
// node should be made "active". Have to find out which node is chosen
// before that can happen.
{
  char line[SLEN];
  int *path;
  int x, y, z;
  boolean nogood;
  dialogue dptr;

  do {
    nogood = false;
    printf("\nSession history:");
    show_session_action(D,0);
    printf("Change to which stage?  ");
    fflush(stdout);
    fgets(line,SLEN,stdin);
    for (x=0, y=1; line[x]; x++)
      if (line[x] == ':')
	y++;
      else if (!isspace(line[x]) && !isdigit(line[x]))
	nogood = true;
    if (!nogood) {
      strcat(line,":");
      path = (int*) malloc(y * sizeof(int));
      for (x=0,z=0; z<y && !nogood; z++) {
	path[z] = 0;
	while (isspace(line[x])) x++;
	while (isdigit(line[x])) {
	  path[z] *= 10;
	  path[z] += (int)(line[x++]-'0');
	}
	while (isspace(line[x])) x++;
	if (line[x] != ':')
	  nogood = true;
	x++;
      }
      if (nogood) {
	free(path);
	path = 0;
      }
    }
  }
  while(nogood);
  if (path[0] != 0)
    nogood = true;
  else {
    for (dptr=D; dptr->parent; dptr=dptr->parent) ;
    for (x=1; x<y && !nogood; x++) {
      nogood = true;
      for (z=0; z<dptr->nsubs; z++)
	if (dptr->sub[z]->relative_node_id == path[x]) {
	  nogood = false;
	  dptr = dptr->sub[z];
	  break;
	}
    }
  }
  if (nogood) {
    printf("\nError: incorrect dialogue path name\n");
    PAUSE();
  }
  else
    active_node = dptr;
  if (path)
    free(path);
}


diagnose_event next_diagnose_event(void)
// Where the core of the dialogue happens: display the menu of options
// and collect the user's choice. Mostly this is one word, but in some
// cases it is two words. Parse it and return the event type.
{
  char line[SLEN];
  char firstword[10];
  char secondword[10];
  char bothwords[20];
  int x, y;
  diagnose_event de;

  system("clear");
  printf("Menu:\n");
  printf("   add  (add a clause)\n");
  printf("   delete  (delete a clause)\n");
  printf("   edit  (change a clause)\n");
  printf("   hard  (make a clause hard)\n");
  printf("   soft  (make a clause soft)\n");
  printf("   weight  (change weight of a soft clause)\n");  
  printf("   time   (adjust time limit\n");

  printf("   model  (solve MAX-CSP)\n");
  printf("   core  (find a minimal unsatisfiable core)\n");
  printf("   optimal  (find the smallest unsatisfiable core)\n");
  printf("   detail  (drill down to lower-level constraints)\n");

  printf("   show problem  (display current .fin file)\n");
  printf("   show model  (re-display most recent model)\n");
  printf("   show core  (re-display most recent MUC)\n");
  printf("   show session  (display dialogue history)\n");

  printf("   save problem  (save current .fin file)\n");
  printf("   save model  (save most recent model)\n");
  printf("   save core  (save most recent MUC)\n");
  printf("   save session  (save dialogue history)\n");

  printf("   revert  (revisit another dialogue node)\n");
  printf("   exit  (quit the session)\n");
  printf("\nSelect:  "); fflush(stdout);
  fgets(line,SLEN,stdin);
  for (x=0; line[x] && isspace(line[x]); x++) ;
  if (!line[x]) return no_op;
  for (y=0; line[x] && isalpha(line[x]); x++) firstword[y++] = line[x];
  firstword[y] = 0;
  while (line[x] && isspace(line[x])) x++;
  for (y=0; line[x] && isalpha(line[x]); x++) secondword[y++] = line[x];
  secondword[y] = 0;
  if (y)
    sprintf(bothwords,"%s %s", firstword, secondword);
  else
    strcpy(bothwords,firstword);
  for (de=add_clause; de<=halt; de++)
    if (!strcmp(bothwords,ename[de]))
      return de;
  return no_op;
}


void process_event(dialogue D, diagnose_event e)
// Mostly a multi-way switch calling the appropriate action for e.
// The event list, a record of the dialogue, is updated here as well.
{
  event_list E;

  if (e != no_op) {
    E = (event_list) malloc(sizeof(EVENT_LIST));
    E->e = e;
    E->container = D->absolute_node_id;
    E->iparams = 0;
    E->sparams[0] = E->sparams[1] = 0;
    E->mresult = 0;
    E->cresult = 0;
    E->sresult = 0;
    E->next = 0;
  }
  switch(e) {
  case add_clause:        add_clause_action(D,E);        break;
  case delete_clause:     delete_clause_action(D,E);     break;
  case edit_clause:       edit_clause_action(D,E);       break;
  case protect_clause:    weight_clause_action(D,E,0);   break;
  case unprotect_clause:  weight_clause_action(D,E,1);   break;
  case weight_clause:     weight_clause_action(D,E,2);   break;
  case change_time_limit: change_time_limit_action(D,E); break;
  case model:             model_action(D,E);             break;
  case core:              core_action(D,E);            break;
  case optcore:           optimal_core_action(D,E);      break;
  case drill_down:        drill_down_action(D,E);        break;
  case show_problem:      show_problem_action(D,E);      break;
  case show_model:        show_model_action(D,E);        break;
  case show_core:         show_core_action(D,E);         break;
  case show_session:      show_session_action(D,E);      break;
  case save_problem:      save_problem_action(D,E);      break;
  case save_model:        save_model_action(D,E);        break;
  case save_core:         save_core_action(D,E);         break;
  case save_session:      save_session_action(D,E);      break;
  default:                                               return;
  }
  if (E->e) {
    if (!D->e) {
      D->e = E;
      E->last = 0;
      E->event_id = 1;
    }
    else {
      for (E->last = D->e; E->last->next; E->last = E->last->next) ;
      E->last->next = E;
      E->event_id = E->last->event_id + 1;
    }
  }
  else free_event_list(E);
}



