/*
 * The main complex computation functions, using FINDER to generate
 * models or to prove that none exist. These are called in order to
 * generate near-models and in order to generate unsatisfiable cores.
 */

#include "diagnose.h"


void model_action(dialogue D, event_list E)
// Generate as good a model as possible, and report it along with a
// list of the violated clauses. Note this prints to stdout.
{
  int good;
  int x, y;

  good = clauses_modelled(D,true,0,1);
  if (!good) {
    iclose(D->theModelFile);
    report_no_model(D);
  }
  else {
    D->theModelFile = iopen();
    xprint_the_model(0,D->theModelFile);
    show_model_action(D,0);
    for (x=1,y=0; x<=D->nclauses; x++)
      if (!D->theClauses[x].satisfied) {
	if (y == 0)
	  printf("Unsatisfied clauses:\n");
	y++;
	printf("%4d%s\n", x, D->theClauses[x].s.c_str);
      }
    if (!y)
      printf("All clauses in the current model version are satisfied.\n");
  }
  FINDER_cleanup();
  PAUSE();
}


boolean clauses_modelled(dialogue D, boolean set_sat, set s, int soft)
// Use local search to find a model. If set_sat then label all clauses
// in D as satisfied or unsatisfied. Return good if all relevant
// clauses are recently satisfied. Softweight parameter is for passing
// to make_finderfile. The subset of clauses to be modelled is s, taken
// to be all clauses if s is not supplied.
{
  ifile basis;
  int good;
  int x;
  int max_fails = 0;

  basis = make_finderfile(D,s,soft);
  iprintf(basis,"setting verbosity job none\n");
  iprintf(basis,"setting verbosity models none\n");
  iprintf(basis,"setting verbosity stats none\n");
  iprintf(basis,"setting time_limit 0\n");
  iprintf(basis,"setting flips %d\n",D->flip);
  iprintf(basis,"setting random %d\n",D->seed);
  iprintf(basis,"setting local_search\n");
  iprintf(basis,"end\n");
  switch(soft) {
  case 2:                          // Called from cosat_finder
    for (x=1; x<=D->nclauses; x++)
      if (s && !S_is_member(x,s))
	max_fails++;
    break;
  case 1:
    if (s)                         // Called from tester
      max_fails = 0;
    else                           // Called from model_action
      max_fails = BIGINT;
  }

  good = finder_test(basis,D,max_fails);
  if (good) {
    if (set_sat) {
      for (x=1; x<=D->nclauses; x++)
	if (!s || S_is_member(x,s)) {
	  D->theClauses[x].satisfied = D->theClauses[x].recently_satisfied;
	  if (D->theClauses[x].sacred && !D->theClauses[x].satisfied)
	    good = 0;
	}
    }
    else {
      for (x=1; x<=D->nclauses; x++)
	if (!s || S_is_member(x,s)) {
	  if (!D->theClauses[x].recently_satisfied)
	    good = 0;
	}
    }
  }
  return good;
}


int finder_test(ifile finfil, dialogue D, int tolerated)
{
  // Use search (init_finder) to see whether the newly added clause(s)
  // can be satisfied. Then if a model was found, label the clauses as
  // "recently satisfied" or not. Return yes or no according to the
  // init_finder call.
  int good;
  dynstring clause_input;
  int x;
  ifile foutfil = iopen();               // Dummy channel to suppress output

  good = initFinder(finfil,foutfil,tolerated);
  iclose(foutfil);
  if (good) {
    for (x=1; x<=D->nclauses; x++) {
      dynstring_initialise(&clause_input,"clause { ");
      dynstring_strcat(&clause_input,&(D->theClauses[x].s));
      dynstring_strcatc(&clause_input," }");
      D->theClauses[x].recently_satisfied = found(&clause_input,0,0);
    }
  }
  return good;
}


void report_no_model(dialogue D)
// Action called when no model of any sort was returned. Prints to stdout.
{
  int sacred_cows = 0;
  int x;

  for (x=1; x<=D->nclauses; x++)
    D->theClauses[x].satisfied = false;
  for (x=1; x<=D->nclauses; x++)
    if (D->theClauses[x].sacred)
      sacred_cows++;
  if (!sacred_cows) {
    printf("No model of any sort was found. ");
    printf("Either pre-processing took too long,\n");
    printf("or the specification of sorts and functions is inconsistent.\n");
    printf("This needs to be fixed before we can examine the clauses.\n");
  }
  else {
    printf("No model at all was found satisfying ");
    if (sacred_cows == 1) printf("the clause ");
    else if (sacred_cows == 2) printf("both of the clauses ");
    else printf("all %d of the clauses ", sacred_cows);
    printf("\nfixed as \"definitely true\".\n");
  }
}


int core_exists(dialogue D)
{
  ifile f, g;
  int good;

  f = make_finderfile(D,0,0);
  g = iopen();
  iprintf(f,"setting verbosity {job:none models:none stats:none}\nend\n");
  good = initFinder(f,g,0);
  iclose(g);
  if (good) {
    printf("\nNo core exists: the clause set is satisfiable\n");
    return 0;
  }
  else
    return 1;
}


void optimal_core_action(dialogue D, event_list E)
{
  if (core_exists(D)) {
    find_optimal_core(D);
    show_core_action(D,E);
  }
}


void find_optimal_core(dialogue D)
{
  int x;

  NO_CORE = false;
  init_init_finder();
  D_initial();
  if (D->core_only)
    D_set_verbosity(0);
  else
    D_set_verbosity(1);
  for (x=1; x<=D->nclauses; x++)
    D_add_element(x,1);
  D->theCore = D_OPT(D,tester,cosat_finder);
}


int tester(D_problem p, set s)
// Call FINDER to test whether the set s of clauses is satisfiable or not.
// Passed as a parameter to D_OPT. 
//
// Need to check what is happening to the input file now that an ifile
// is being used. Maybe some more copying of the basis is in order.
{
  ifile basis;
  int good;
  int x;

  D_set_verbosity(0);

  basis = make_finderfile(p,s,0);
  good = complete_search(p,basis,1,0,0);
  if (!good) return 1;
  if (good == 1) return 0;
  // No model found, but no proof there is none.
  p->flip *= 2;
  good = clauses_modelled(p,false,s,1);
  p->flip /= 2;
  FINDER_cleanup();
  if (good == 1) return 0;
  // Still no model found by either search method
  basis = make_finderfile(p,s,2);
  good = complete_search(p,basis,2,1,BIGINT);
  if (!good) return 1;
  if (good == 1) return 0;
  // No model found, but STILL no proof there is none.
  basis = make_finderfile(p,s,0);
  good = complete_search(p,basis,5,2,0);
  if (!good) return 1;
  if (good == 1) return 0;
  printf("\nPossible core:\n");
  for (x=1; x<=p->nclauses; x++)
    if (S_is_member(x,s))
      printf("%4d%s\n", x, p->theClauses[x].s.c_str);
  printf("\nThis has not been shown unsatisfiable, but no model was found\n");
  printf("within an extended time limit of %d seconds.\n",p->time*5);
  PAUSE();
  NO_CORE = true;
  return 1;
}


int complete_search(D_problem p, ifile basis,
		    int factor, int increment, int tolerated)
// Look for a model of the clauses in basis. If they are shown to be
// satisfiable, return 1. If they are shown to be unsatisfiable,
// return 0. If no decision was reached because of timeout (or
// insertion limit) return -1.
{
  boolean good;
  char stopcode[SLEN];

  iprintf(basis,"setting verbosity job none\n");
  iprintf(basis,"setting verbosity models none\n");
  iprintf(basis,"setting verbosity stats none\n");
  iprintf(basis,"setting solutions 1\n");
  iprintf(basis,"setting time_limit 0\n");
  iprintf(basis,"setting insertions %d\n",p->time*factor*INSERT_FACTOR);
  iprintf(basis,"setting random %d\n",p->seed+increment);
  iprintf(basis,"setting silent_breaks\n");
  iprintf(basis,"end\n");
  good = finder_test(basis,p,tolerated);
  stop_code(stopcode);
  FINDER_cleanup();
  if (good) return 1;
  if (!strcmp(stopcode,"exhaustion")) return 0;
  return -1;
}


set cosat_finder(D_problem p, set s)
// Call FINDER to use DFBB to find a (near) max-CSP solution to D.
// The set s must be satisfied. It is known to be satisfiable.
// Return the set of unsatisfied clauses.
{
  ifile basis;
  set next_diagnosis = S_new(p->nclauses,0);
  set the_diagnosis = S_new(p->nclauses,0);
  int x;
  int verdict;
  set splus = 0;

  basis = make_finderfile(p,s,2);
  verdict = complete_search(p,basis,1,0,BIGINT);
  for (x=1; x<=p->nclauses; x++) {
    if (!p->theClauses[x].recently_satisfied) {
      if (S_is_member(x,s))
	abort_diagnose("Member of s not recently satisfied",0);
      S_into(x,next_diagnosis);
    }
  }
  if (verdict != 1) {
    //    p->flip *= 3;
    verdict = (int) clauses_modelled(p,false,s,2);
    //    p->flip /= 3;
    FINDER_cleanup();
    if (verdict) {
      S_empty(next_diagnosis);
      for (x=1; x<=p->nclauses; x++) {
	if (!p->theClauses[x].recently_satisfied) {
	  if (S_is_member(x,s))
	    abort_diagnose("Member of s not recently satisfied",0);
	  S_into(x,next_diagnosis);
	}
      }
    }
    else {
      splus = S_copy(s,0);
      Forallmembers(s,x) {
	if (!(p->theClauses[x].recently_satisfied))
	  abort_diagnose("Member of s has become unsatisfied\n",0);
      }
      for (x=1; x<=p->nclauses; x++) {
	S_into(x,splus);
	if (p->theClauses[x].recently_satisfied)
	  S_outof(x,next_diagnosis);
	else {
	  basis = make_finderfile(p,splus,0);
	  verdict = complete_search(p,basis,1,0,0);
	  if (verdict == 1)
	    S_outof(x,next_diagnosis);
	  else {
	    S_into(x,next_diagnosis);
	    S_outof(x,splus);
	  }
	}
      }
    }
  }
  // Make a copy of the diagnosis, so that it has no ex-members
  Forallmembers(next_diagnosis,x)
    S_into(x,the_diagnosis);
  S_free(next_diagnosis);
  S_free(splus);
  return the_diagnosis;
}


void core_action(dialogue D, event_list E)
{
  if (core_exists(D)) {
    make_non_optimal_core(D);
    show_core_action(D,E);
  }
}


void make_non_optimal_core(dialogue D)
// Seek a core (any core) and copy it to D->theCore. If there is no
// decision as to whether certain clauses are in the core, copy them
// to D->theNearCore.
{
  set C = S_new(D->nclauses,0);
  int x;
  int result;
  ifile f;

  NO_CORE = false;
  D->dunno = S_new(D->nclauses,0);
  init_init_finder();
  for (x=1; x<=D->nclauses; x++)
    S_into(x,C);
  for (x=1; x<=D->nclauses; x++) {
    S_outof(x,C);
    f = make_finderfile(D,C,0);
    result = complete_search(D,f,1,0,0);
    if (result)
      S_into(x,C);         // May be required for the core
    if (result < 0)
      S_into(x,D->dunno);  // May not be required
  }
  if (!D->core_only) {
    system("clear");
    if (S_size(D->dunno)) {
      if (S_size(D->dunno) < S_size(C)) {
	printf("The following are definitely in the core:\n");
	Forallmembers(C,x) {
	  if (!S_is_member(x,D->dunno))
	    printf("  %-4d %s\n", x, D->theClauses[x].s.c_str);
	}
	printf("\nThe following may be in the core:\n");
	Forallmembers(D->dunno,x)
	  printf("  %-4d %s\n", x, D->theClauses[x].s.c_str);
      }
      else
	printf("No definite core found within the time limit.\n");
      printf("If you want me to try harder, please increase the time limit.");
      PAUSE();
    }
  }
  D->theCore = S_copy(C,D->theCore);
  if (S_size(D->dunno) == 0) {
    S_free(D->dunno);
    D->dunno = 0;
  }
}
