/*
 * Functions to save the problem, model, core and session. Also
 * functions to show the same to the user. Function to prompt for a
 * filename (for save purposes) is in here.
 */

#include "diagnose.h"


void show_problem_action(dialogue D, event_list E)
{
  int x;

  system("clear");
  for (x=1; x<=D->nclauses; x++) {
    if (D->theClauses[x].softweight == 0)
      printf("hard ");
    else if (D->theClauses[x].softweight == 10)
      printf("firm ");
    else
      printf("soft ");
    printf("%4d%s\n", x, D->theClauses[x].s.c_str);
  }
  if (E) PAUSE();
}


void show_core_action(dialogue D, event_list E)
{
  if (D->core_only) {
    if (D->theCore)
      print_core(D);
    else
      printf("No core: the clause set is satisfiable\n");
  }
  else {
    system("clear");
    if (D->theCore)
      print_core(D);
    else
      printf("No unsatisfiable core [yet].\n");
    PAUSE();
  }
}


void print_core(dialogue D)
{
  int x;

  for (x=1; x<=D->nclauses; x++)
    if (S_is_member(x,D->theCore)) {
      printf("%4d%s\n", x, D->theClauses[x].s.c_str);
      if (D->dunno && S_is_member(x,D->dunno))
	printf("  (may be redundant)\n");
    }
  if (D->dunno && S_size(D->dunno) > 0) {
    printf("\nTo reduce uncertainty, increase the time limit. If the limit\n");
    printf("is already maximal, sorry: this is as good as it gets!\n");
  }
}


void show_model_action(dialogue D, event_list E)
{
  if (!D->theModelFile)
    printf("No model is recorded for the current version of the problem.\n");
  else
    dumpitof(D->theModelFile,stdout);
  if (E) PAUSE();
}


void show_session_action(dialogue D, event_list E)
{
  dialogue root;

  printf("\n");
  for (root=D; root->parent; root=root->parent) ;
  show_local_session(root,stdout);
  printf("\nCurrent dialogue stage is ");
  print_relative_id(D,stdout);
  printf("\n");
  if (E) PAUSE();
}

void show_local_session(dialogue D, FILE *f)
{
  event_list eptr;
  int x;

  for (eptr=D->e; eptr; eptr=eptr->next)
    if (eptr->e == no_op)
      show_local_session(D->sub[eptr->iparams[0]],f);
    else {
    //    fprintf(f,"%3d.%-3d ",D->absolute_node_id,eptr->event_id);
    print_relative_id(D,f);
    printf(" [%d]",eptr->event_id);
    if (eptr->event_id < 10) printf(" ");
    fprintf(f," %s",ename[eptr->e]);
    for (x=0; x<niparams(eptr->e); x++)
      fprintf(f," %d",eptr->iparams[x]);
    if (eptr->sparams[0])
      fprintf(f," %s",eptr->sparams[0]);
    if (eptr->sparams[1])
      fprintf(f," %s",eptr->sparams[1]);
    fprintf(f,"\n");
  }
}

void print_relative_id(dialogue D, FILE *f)
{
  if (D->parent)
    print_relative_id(D->parent,f);
  fprintf(f,"%c%d",(D->parent? ':': ' '),D->relative_node_id);
}


void prompt_for_filename(char *fname)
// Just what you would expect. Note that spaces are not allowed in
// filenames. The space removal also cleans leading spaces and 
// trailing line feeds and suchlike dross.
{
  int x, y;

  printf("Save as: "); fflush(stdout);
  fgets(fname,SLEN,stdin);
  for (x=0,y=0; fname[x]; x++)
    if (!isspace(fname[x]))
      fname[y++] = fname[x];
  fname[y] = 0;  
}


void save_problem_action(dialogue D, event_list e)
// Save the current problem to a finder input file named by the user.
// Record the file name in the string parameters of the event.
{
  char fname[SLEN];
  FILE *outfil;
  ifile infil;

  prompt_for_filename(fname);
  outfil = fopen(fname,"w");
  if (!outfil)
    printf("Sorry: could not open file %s for writing.\n",fname);
  else {
    infil = make_finderfile(D,0,1);
    iprintf(infil,"end.\n");
    dumpitof(infil,outfil);
    iclose(infil);
    fclose(outfil);
    e->sparams[0] = (char*) malloc(strlen(fname));
    strcpy(e->sparams[0],fname);
  }
}


void save_model_action(dialogue D, event_list e)
{
  char fname[SLEN];
  FILE *outfil;

  if (D->theModelFile) {
    prompt_for_filename(fname);
    outfil = fopen(fname,"w");
    if (outfil) {
      dumpitof(D->theModelFile,outfil);
      fclose(outfil);
      e->sparams[0] = (char*) malloc(strlen(fname));
      strcpy(e->sparams[0],fname);
    }
    else
      printf("Sorry: could not open file %s for writing.\n",fname);
  }
  else
    printf("There is no current model to save\n");
}


void save_core_action(dialogue D, event_list E)
{
  int x;
  char fname[SLEN];
  FILE *outfil;

  if (D->theCore) {
    prompt_for_filename(fname);
    outfil = fopen(fname,"w");
    if (outfil) {
      for (x=1; x<=D->nclauses; x++)
	if (S_is_member(x,D->theCore))
	  fprintf(outfil,"%4d%s\n", x, D->theClauses[x].s.c_str);
      fclose(outfil);
      E->sparams[0] = (char*) malloc(strlen(fname));
      strcpy(E->sparams[0],fname);
    }
    else
      printf("Sorry: could not open file %s for writing.\n",fname);
  }
  else
    printf("No unsatisfiable core yet.\n");
}


void save_session_action(dialogue D, event_list e)
// Save the session so far to a text file named by the user.
// Record the file name in the string parameters of the event.
{
  dialogue root;
  char fname[SLEN];
  FILE *outfil;

  prompt_for_filename(fname);
  outfil = fopen(fname,"w");
  if (outfil) {
    for (root=D; root->parent; root=root->parent) ;
    show_local_session(root,outfil);
    fclose(outfil);
    e->sparams[0] = (char*) malloc(strlen(fname));
    strcpy(e->sparams[0],fname);
  }
  else
    printf("Sorry: could not open file %s for writing.\n",fname);
}
