/*
 * Mostly low-level functions carrying out utility functions. Includes
 * basic clause manipulations such as copying and deleting
 * clauses. Also the functions to convert the initial input into a
 * more usable form and to make a finder file are in here.
 */

#include "diagnose.h"

static ifile SandF = 0;


void PAUSE(void)
// Prompt for RETURN and hang until the user obliges
{
  char s[SLEN];

  printf("\nRETURN to return to the menu");
  fflush(stdout);
  fgets(s,SLEN,stdin);
}


void copy_dclause(dclause source, dclause dest)
// Copy the contents of source to dest. Dest must have been allocated.
// Since the number and length of strings may change, some freeing of
// memory in dest may be required first.
{
  int x;

  if (!source || source == dest) return;
  if (!dest)
    abort_diagnose("Attempt to copy a clause to null",0);
  dclause_init(dest);
  dest->var_name = (char**) malloc(source->var_count * sizeof(char*));
  dest->var_sort = (char**) malloc(source->var_count * sizeof(char*));
  for (x=0; x<source->var_count; x++) {
    dest->var_name[x] = (char*) malloc(strlen(source->var_name[x]));
    strcpy(dest->var_name[x],source->var_name[x]);
    dest->var_sort[x] = (char*) malloc(strlen(source->var_sort[x]));
    strcpy(dest->var_sort[x],source->var_sort[x]);
  }
  dest->id = source->id;
  dest->pid = source->pid;
  dest->weight = source->weight;
  dest->var_count = source->var_count;
  dest->tuples = source->tuples;
  dest->softweight = source->softweight;
  dest->sacred = source->sacred;
  dest->satisfied = source->satisfied;
  dest->recently_satisfied = source->recently_satisfied;
  dest->flip = source->flip;
  dest->cosat = source->cosat;
  dynstring_strcpy(&(dest->s),&(source->s));
}


void problem_basis(ifile source)
// Search source file for sorts and functions and copy them into the
// ifile SandF. The source file is expected to be in the format produced
// by FINDER when dumping the job. Arbitrary FINDER input will not be
// recognised by this function, so it must be parsed and dumped
// first. The utility function transfer_job should have been called to
// do this if the input was initially in user-generated format.
{
  char line[1000];
  char keyword[100];
  int x;

  ireread(source);                         // Sets input point to start of file
  SandF = ifile_init(SandF);
  while (igets(line,1000,source)) {
    for (x=0; line[x] && isspace(line[x]); x++) ;
    if (line[x]) {
      sscanf(line,"%s",keyword);
      if (!strcmp(keyword,"sort") ||
	  !strcmp(keyword,"function") ||
	  !strcmp(keyword,"vocabulary")) {
	iputs(line,SandF);
	while (!strchr(line,'}')) {
	  igets(line,1000,source);
	  iputs(line,SandF);
	}
      }
    }
  }
}


dclause problem_clauses(ifile fin, int *nclauses)
// Scan fin for clauses and make them into local dclauses
{
  char line[10000];
  char keyword[100];
  int x;
  dclause allClauses = 0;

  *nclauses = 0;
  ireread(fin);
  while (igets(line,10000,fin)) {
    for (x=0; line[x] && isspace(line[x]); x++) ;
    if (line[x]) {
      sscanf(line,"%s",keyword);
      if (!strcmp(keyword,"clause") || !strcmp(keyword,"constraint")) {
	(*nclauses)++;
	if ((*nclauses)%10 == 1)
	  allClauses = (dclause) realloc(allClauses,
					 ((*nclauses)+10)*sizeof(DCLAUSE));
	get_dclause(line,allClauses+(*nclauses));
      }
    }
  }
  return allClauses;
}


dclause get_dclause(char *line, dclause dest)
// Make the clause on line into a dclause and put it in in dest
{
  int x;
  int y = 0;
  int z;
  char *temp;

  temp = (char*) malloc(strlen(line));
  for (x=0; line[x] && line[x] != '{'; x++) ;
  for (x++; line[x] && line[x] != '}'; x++) temp[y++] = line[x];
  temp[y] = 0;
  for (; y >= 0 && temp[y] != '.'; y--) temp[y] = 0;
  while (line[x] != '%') x++; x++;
  if (!dest) dest = (dclause) malloc(sizeof(DCLAUSE));
  dclause_init(dest);
  sscanf(line+x,"%d%d%d%d%d%d",
	 &(dest->var_count),
	 &(dest->weight),
	 &(dest->tuples),
	 &(dest->softweight),
	 &(dest->id),
	 &(dest->pid));
  for (y=0; y<6; y++) {
    while (line[x] && isspace(line[x])) x++;
    while (line[x] && !isspace(line[x])) x++;
  }
  if (dest->var_count) {
    dest->var_name = (char**) malloc(sizeof(char*)*dest->var_count);
    dest->var_sort = (char**) malloc(sizeof(char*)*dest->var_count);
    for (y=0; y<dest->var_count; y++) {
      while (isspace(line[x])) x++;
      for (z=0; !isspace(line[x+z]); z++);
      dest->var_name[y] = (char*) malloc(sizeof(char)*(z+1));
      sscanf(line+x,"%s",dest->var_name[y]);
      while (!isspace(line[x])) x++;
      while (isspace(line[x])) x++;
      for (z=0; !isspace(line[x+z]); z++);
      dest->var_sort[y] = (char*) malloc(sizeof(char)*(z+1));
      sscanf(line+x,"%s",dest->var_sort[y]);
      while (!isspace(line[x])) x++;
    }
  }
  dynstring_initialise(&(dest->s),temp);
  if (dest->softweight)
    dest->softweight = 1;
  else
    dest->softweight = 10;
  dest->sacred = false;
  dest->satisfied = dest->recently_satisfied = false;
  dest->flip = dest->cosat = 0;
  return dest;
}


int niparams(diagnose_event e)
// Number of integer parameters of events of type e
{
  switch(e) {
  case no_op:
  case add_clause:
  case delete_clause:
  case edit_clause:
  case change_time_limit:
  case revert: return 1;
  case protect_clause:
  case unprotect_clause:
  case weight_clause: return 2;
  default: return 0;
  }
}


void set_up_event_names(void)
// Associate strings with the event name pointers
{
  ename[no_op] = "";
  ename[add_clause] = "add";
  ename[delete_clause] = "delete";
  ename[edit_clause] = "edit";
  ename[protect_clause] = "hard";
  ename[unprotect_clause] = "soft";
  ename[weight_clause] = "weight";
  ename[change_time_limit] = "time";
  ename[model] = "model";
  ename[core] = "core";
  ename[optcore] = "optimal";
  ename[drill_down] = "detail";
  ename[show_problem] = "show problem";
  ename[show_session] = "show session";
  ename[show_model] = "show model";
  ename[show_core] = "show core";
  ename[save_model] = "save model";
  ename[save_core] = "save core";
  ename[save_problem] = "save problem";
  ename[save_session] = "save session";
  ename[revert] = "revert";
  ename[halt] = "exit";
}


step_direction direction(diagnose_event e)
// The dialogue direction (forward, back, stay) of the event e
{
  switch(e) {
  case revert:
  case halt:
    return back;
  case protect_clause:
  case unprotect_clause:
  case weight_clause:
  case add_clause:
  case delete_clause:
  case edit_clause:
    return forward;
  default:
    return stay;
  }
}


ifile make_finderfile(dialogue D, set s, int soft)
// Make the current problem into a finder input file. The set s is an
// optional mask to determine which clauses are included. No settings
// are specified. The "end" line is not appended, and the open ifile
// is returned so that more lines can be added.
{
  int x;
  char line[SLEN];
  ifile outfil = iopen();

  ireread(SandF);
  while (igets(line,SLEN,SandF))
    iputs(line,outfil);
  for (x=1; x<=D->nclauses; x++)
    switch(soft) {
    case 0:                          // Hard constraints only
      if (!s || S_is_member(x,s))
	iprintf(outfil,"clause { %s }\n",D->theClauses[x].s.c_str);
      break;
    case 1:                          // Soft weights as given
      iprintf(outfil,"clause (%d) { %s }\n",
	      D->theClauses[x].softweight,
	      D->theClauses[x].s.c_str);
      break;
    case 2:                          // Set s hard, all others soft
      if (s && S_is_member(x,s))
	iprintf(outfil,"clause { %s }\n",D->theClauses[x].s.c_str);
      else
	iprintf(outfil,"clause (1) { %s }\n",D->theClauses[x].s.c_str);
      break;
    default:
      abort_diagnose("Softweight parameter out of range",0);
    }
  return outfil;
}


void dclause_init(dclause c)
// Assuming c exists, zero its fields
{
  dynstring_initialise(&(c->s),"");
  c->var_name = c->var_sort = 0;
  c->var_count = 0;
  c->id = c->pid = 0;
  c->weight = c->tuples = c->softweight = 0;
  c->sacred = c->satisfied = c->recently_satisfied = c->flip = false;
  c->cosat = 0;
}

void dclause_free(dclause c)
// Free the allocated parts of c but don't try to free c itself
{
  int x;

  dynstring_free(&(c->s));
  if (c->var_name) {
    for (x=0; x<c->var_count; x++) {
      if (c->var_name[x])
	free(c->var_name[x]);
      c->var_name[x] = 0;
    }
    free(c->var_name);
  }
  if (c->var_sort) {
    for (x=0; x<c->var_count; x++)
      if (c->var_sort[x]) {
	free(c->var_sort[x]);
	c->var_sort[x] = 0;
      }
    free(c->var_sort);
  }
  c->var_count = 0;
  c->var_name = c->var_sort = 0;   // Prevent double free
}


void free_event_list(event_list E)
// Free all the parts of E and then free E itself.
// Also resets everything to zero. This is not necessary but may help
// during debugging with gdb or the like.
{
  if (E) {
    E->container = 0;
    E->event_id = 0;
    E->e = 0;
    S_free(E->mresult);                      E->mresult = 0;
    S_free(E->cresult);                      E->cresult = 0;
    if (E->iparams) free(E->iparams);        E->iparams = 0;
    if (E->sparams[0]) free(E->sparams[0]);  E->sparams[0] = 0;
    if (E->sparams[1]) free(E->sparams[1]);  E->sparams[1] = 0;
    if (E->sresult) free(E->sresult);        E->sresult = 0;   
    if (E->last) E->last->next = E->next;
    if (E->next) E->next->last = E->last;
    E->last = E->next = 0;
    free(E);
  }
}


void free_the_clauses(dialogue D)
// Pretty much what it says: throw away all the clauses in D and
// free the memory they used. Purpose is either to clean up D on
// exit from the dialogue tree, or to prepare to replace the clauses
// by an updated version.
// As in the other free routines, the variables freed are then zeroed
// to make debugging with gdb easier.
{
  int x, y;

  for (x=0; x<=D->nclauses; x++)
    if (D->theClauses[x].var_count) {
      for (y=0; y<D->theClauses[x].var_count; y++) {
	free(D->theClauses[x].var_name[y]);  D->theClauses[x].var_name[y] = 0;
	free(D->theClauses[x].var_sort[y]);  D->theClauses[x].var_sort[y] = 0;
      }
      D->theClauses[x].var_count = 0;
      free(D->theClauses[x].var_name);  D->theClauses[x].var_name = 0;
      free(D->theClauses[x].var_sort);  D->theClauses[x].var_sort = 0;
    }
  free(D->theClauses);
  D->theClauses = 0;
  D->nclauses = 0;
}


void free_dialogue_tree(dialogue D)
// Recursively free fields of D and D itself
{
  int x;

  if (!D) return;
  for (x=0; x<D->nsubs; x++) {
    free_dialogue_tree(D->sub[x]);
    D->sub[x] = 0;
  }
  free(D->sub);
  D->nsubs = 0;
  free_the_clauses(D);
  S_free(D->theCore);
  D->theCore = 0;
  D->absolute_node_id = D->relative_node_id = 0;
  if (D->e) {
    while (D->e->next)
      free_event_list(D->e->next);
    free_event_list(D->e);
    D->e = 0;
  }
  D->seed = D->time = D->flip = 0;
  D->parent = 0;
  free(D);
}


void abort_diagnose(char *s1, char *s2)
{
  printf("\nAborting diagnose: %s",s1);
  if (s2) printf(" %s",s2);
  printf("\n");
  fflush(stdout);       // Just in case...
  exit(1);
}
