#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <unistd.h>
#include "dynstring.h"
#include "dual.h"
#ifndef IFILES
#include "ifile.h"
#endif

#define SLEN 1000
#define defaultTimelimit 2
#define defaultFliplimit 15000
#define theLowFliplimit 10000
#define defaultSeed 1
#define BIGINT 100000000
#define insert_factor_default 200000
#define DEFAULT_WIDTH 1000
#define DEFAULT_HEIGHT 800


typedef enum {
  no_op,              // The null (no valid action selected)
  add_clause,         // New node
  delete_clause,      // New node
  edit_clause,        // New node
  protect_clause,     // Make clause hard. New node
  unprotect_clause,   // Make clause soft. New node
  weight_clause,      // Change weight of soft clause. New node
  change_time_limit,  // Same node
  model,              // Solve max-csp. Same node
  core,               // Solve MUC. Same node
  optcore,            // Solve OMUC. Same node
  drill_down,         // See lower level of core. NYI
  show_problem,       // Display contents of amended .fin file
  show_session,       // Display tree of old nodes. Same node
  show_model,         // Re-display solution to Max-sat. Same node
  show_core,          // Redisplay unsatisfiable core. Same node
  save_model,         // Same node
  save_core,          // Same node
  save_problem,       // Print .fin file as amended. Same node
  save_session,       // Dump whole trace. Same node
  revert,             // Revisit old node
  halt                // Exit from dialogue altogether
} diagnose_event;


typedef enum {
  back,
  stay,
  forward
} step_direction;


typedef struct {
  int id;
  int pid;
  dynstring s;
  int weight;
  int var_count;
  int tuples;
  char **var_name;
  char **var_sort;
  int softweight;
  boolean sacred;
  boolean satisfied;
  boolean recently_satisfied;
  boolean flip;
  int cosat;         /* Number of times in the cosatisfied set */
} DCLAUSE, *dclause;

typedef struct event_list_t {
  int container;                      // Absolute id of its node
  int event_id;                       // Sequence number within node
  diagnose_event e;                   // What just happened
  int *iparams;                       // With what integer parameters
  char *sparams[2];                   // With what string parameters
  set mresult;                        // Unsatisfied clauses in model
  set cresult;                        // Minimal unsatisfiable core
  char *sresult;                      // Any string that helps
  struct event_list_t *last;
  struct event_list_t *next;
} EVENT_LIST, *event_list;

typedef struct dialogue_structure {
  int absolute_node_id;               // Unique identifier
  int relative_node_id;               // Which child of parent
  event_list e;                       // List of events at this node
  int nclauses;                       // How many clauses there are
  dclause theClauses;                 // Local copy of the problem (array)
  int seed;                           // For random number generator
  int time;                           // Time limit for searches
  int flip;                           // Flip limit for local searches
  int nsubs;                          // How many children there are
  set theCore;                        // Unsatisfiable subset of clauses
  set dunno;                          // Subset of theCore, possibly not core
  ifile theModelFile;                 // File containing current model
  int core_only;                      // No user involvement: return a core
  struct dialogue_structure **sub;
  struct dialogue_structure *parent;
} DIALOGUE, *dialogue;

typedef struct list_of_int {
  int i;
  struct list_of_int *last;
  struct list_of_int *next;
} INTLIST, *intlist;


#ifdef DIAGNOSE_MAIN
#define D_GLOBAL
#else
#define D_GLOBAL extern
#endif

D_GLOBAL dialogue active_node;       // Points to where we are in the tree
D_GLOBAL char *ename[halt+1];        // Names of event types
D_GLOBAL boolean NO_CORE;            // Search for a core timed out
D_GLOBAL int INSERT_FACTOR;          // Expected inserts per second (adjustable)


// DIALOGUE

dialogue get_dialogue_node(int verbose);
dialogue set_up_diagnose(int argc, char *argv[]);
dialogue do_new_dialogue(dialogue p, diagnose_event cause);
void do_dialogue(dialogue D);
void new_active_node(dialogue D);
diagnose_event next_diagnose_event(void);
void process_event(dialogue D, diagnose_event e);


// UTILITIES

void PAUSE(void);
// Prompt for RETURN and hang until the user obliges

void copy_dclause(dclause source, dclause dest);
// Copy the contents of source to dest. Dest must have been allocated.

void problem_basis(ifile source);
// Search source file for sorts and functions and copy them into the
// ifile SandF. The source file is expected to be in the format produced
// by FINDER when dumping the job. Arbitrary FINDER input will not be
// recognised by this function, so it must be parsed and dumped
// first. The utility function transfer_job should have been called to
// do this if the input was initially in user-generated format.

dclause problem_clauses(ifile fin, int *nclauses);
// Scan fin for clauses and make them into local dclauses

dclause get_dclause(char *line, dclause dest);
// Make the clause on line into a dclause and put it in in dest

int niparams(diagnose_event e);
// Number of integer parameters of events of type e

void set_up_event_names(void);
// Associate strings with the event name pointers

step_direction direction(diagnose_event e);
// The dialogue direction (forward, back, stay) of the event e

ifile make_finderfile(dialogue D, set s, int soft);
// Make the current problem into a finder input file. The set s is an
// optional mask to determine which clauses are included. No settings
// are specified. The "end" line is not appended, and the open ifile
// is returned so that more lines can be added.

void dclause_init(dclause c);
// Assuming c exists, zero its fields

void dclause_free(dclause c);
// Free the allocated parts of c but don't try to free c itself

void free_event_list(event_list E);
// Free all the parts of E and then free E itself.

void free_the_clauses(dialogue D);
// Throw away all clauses in D and free the memory they used

void free_dialogue_tree(dialogue D);
// Recursively free fields of D and D itself

void abort_diagnose(char *s1, char *s2);


// SHOW AND SAVE

void show_problem_action(dialogue D, event_list E);
void show_core_action(dialogue D, event_list E);
void print_core(dialogue D);
void show_model_action(dialogue D, event_list E);
void show_session_action(dialogue D, event_list E);
void show_local_session(dialogue D, FILE *f);
void print_relative_id(dialogue D, FILE *f);
void prompt_for_filename(char *fname);
void save_problem_action(dialogue D, event_list e);
void save_model_action(dialogue D, event_list e);
void save_core_action(dialogue D, event_list e);
void save_session_action(dialogue D, event_list e);


// CHANGE PROBLEM

void add_clause_action(dialogue D, event_list E);
void edit_clause_action(dialogue D, event_list E);
void delete_clause_action(dialogue D, event_list E);
void weight_clause_action(dialogue D, event_list E, int w);
void change_time_limit_action(dialogue D, event_list E);
void drill_down_action(dialogue D, event_list E);
int no_error(ifile f);


// GRUNT

void model_action(dialogue D, event_list E);
boolean clauses_modelled(dialogue D, boolean set_sat, set s, int soft);
int finder_test(ifile finfil, dialogue D, int tolerated);
void report_no_model(dialogue D);
void optimal_core_action(dialogue D, event_list E);
void find_optimal_core(dialogue D);
int complete_search(D_problem p, ifile basis,
		    int factor, int increment, int tolerated);
int tester(D_problem p, set s);
set cosat_finder(D_problem p, set s);
void core_action(dialogue D, event_list E);
void make_non_optimal_core(dialogue D);



/*------ The following are in file FINDER.c or futils.c----------------*/
void get_options(int argc, char *argv[],
		 boolean *debugging,
		 char *filename,
		 char *diag_output,
		 boolean *pipe_out
		 );

int found( const dynstring *s, int add_if_true, int seek_model );
boolean clause_is_long(void);
void init_init_finder(void);
//boolean init_finder(char filename[], int tolerated);
boolean initFinder(ifile i, ifile o, int tolerated);
ifile transfer_job(char *source);
void go(int x);
void refute_current(void);
int theory_size(int x);
void finder_input( const dynstring *s );
void short_timesup(int x);
boolean initial_model_found(void);
void set_no_semantics(void);
int no_semantics_set(void);
void set_job_only(void);
void unset_job_only(void);
void FINDER_cleanup(void);
void print_model_anyway(void);
void stop_code(char *dest);

void inerror(char *s1, char *s2);
boolean too_long_to_be_soft();

/*------ The following is in file foutput.c -----------------------------*/

void print_the_model(void);
void fprint_the_model(FILE *fil);
void iprint_the_model(ifile ifil);
void xprint_the_model(FILE *fil, ifile ifil);

