/*------------------------------------------------------------------------*/
/* Main routines for obtaining finder input */
/*------------------------------------------------------------------------*/

#include "FINDER.h"

static dynstring job_line;    /* Line of input text		     */
static FILE *source;          /* Current input stream if external    */
static ifile isource;         /* Current input stream if internal    */
static boolean no_increment;  /* Don't increment line counter at EOL */


/*------------------------------------------------------------------------*/
char next_char(void)

     /*Called by get_a_char() and get_word().
      *Returns the next character from the job_line[]
      *Skipping over comments and if job_line[] is exhausted, reads
      *the next input line.
      */

{
  while (job_line.c_str[input_point] == '%' ||  /* skip comments    */
	 job_line.c_str[input_point] == '\0') { /* skip empty lines */
    if (source &&                       /* if reading from a file */
	!dynstring_fgets(&job_line,source)) /* refill the job_line */
      return '\0';                 /* patch for end_of_file */
    if (isource &&
	!dynstring_igets(&job_line,isource))
      return '\0';
    input_point = 0;               /* restart at the begining of a line */
    if (!no_increment)
      current_line++;              /* have moved to the next line */
    if (job_line.c_str[0] == '#') {
      no_increment = true;         /* Line starts with # so no increment */
      input_point++;
    }
    else
      no_increment = false;
  }

  return job_line.c_str[input_point++]; /* next charater on the job_line */

} /* next_char() */



/*------------------------------------------------------------------------*/
char get_a_char(void)

     /*Called by get_word().
      *Returns next meaningful character from the job_line*/
     
{
  unsigned char c;

  while (isspace(c = next_char())) ;
  return c;

} /* get_a_char() */



/*------------------------------------------------------------------------*/
void get_word(dynstring *word)
     /*Called by get_key_word().
      *most fcn's in this file and by fcn's in parser.c;
      
      *Gets next meaningful word from the job_line*/
{
  int i;

  dynstring_ensure_space(word, 1);

  word->c_str[0] = get_a_char();

  if (strchr("=><+,;:.%)(}{",word->c_str[0]))
    word->c_str[1] = '\0';
  else {
    for (i=1; !(isspace(word->c_str[i] = next_char())); i++) {
      dynstring_ensure_space(word, i + 1);
      if (i==1 && (!strncmp(word->c_str,"->",2) ||
		   !strncmp(word->c_str,"!=",2))) {
	i = 2;
	break;
      }
      else if ((i==1 && word->c_str[0]=='-') ||
	       strchr("=><+-,;:.%#)(}{",word->c_str[i])) {
	input_point--;
	break;
      }
    }
    word->c_str[i] = '\0';
    word->length = strlen(word->c_str);
  }
} /* get_word() */



/*------------------------------------------------------------------------*/
void input_to_terminator(char *s)

     /* Append input characters to the job_line until a character from
      * s turns up. Incidentally turn newlines into spaces */
{
  char c;
  int i;
  dynstring dummy_line;

  dynstring_initialise(&dummy_line,"");
  for (i=0; job_line.c_str[i]; i++)
    if (isspace(job_line.c_str[i])) job_line.c_str[i] = ' ';
  do {
    c = fgetc(source);
    if (c == '%')
      dynstring_fgets(&dummy_line,source);
    else {
      if (isspace(c)) c = ' ';
      dynstring_append_char(&job_line, c);
    }
  }
  while (!strchr(s,c));
}




/*------------------------------------------------------------------------*/
statement_type get_key_word(void)

     /*Called by read_job().
      *Gets next key word from the job_line and determines its type.*/
     
{
  dynstring wd;
  statement_type result = nul;

  dynstring_initialise(&wd, "");
  
  get_word(&wd);
  if (!strcmp(wd.c_str,".") || !strcmp(wd.c_str,"end")) result = nul;
  else if (!strcmp(wd.c_str,"sort")) result = srt;
  else if (!strcmp(wd.c_str,"function")) result = fcn;
  else if (!strcmp(wd.c_str,"constant")) result = fcn0;
  else if (!strcmp(wd.c_str,"vocabulary")) result = vcb;
  else if (!strcmp(wd.c_str,"clause") ||
	   !strcmp(wd.c_str,"constraint")) result = cls;
  else if (!strcmp(wd.c_str,"setting")) result = prm;
  else if (!strcmp(wd.c_str,"include")) result = inc;
  else inerror (wd.c_str,
		"found where key word (sort/function/vocabulary/clause/setting/end) expected",0);
  dynstring_free(&wd);
  return result;
} /* get_key_word() */



/*------------------------------------------------------------------------*/
void read_job(FILE *infil, ifile ifil)
     
     /*Called by FINDER fcn main() and SCOTT fcn model_finder_file() .
      *Loops continuously over input file until get_key_word() returns null.
      *Integrates the object according to its type*/
{
  if (cmd_line_input) {
    input_point = 0;
    source = 0;
    dynstring_strcpy(&job_line,&command_line);
    dynstring_strcatc(&job_line, " end.");
  }
  else
    dynstring_clear(&job_line);
  input_point = 0;
  source = infil;
  isource = ifil;
  if (source && isource)
    skipout("Input stream is internal and external",SKIP);
  for (;;)
    switch(get_key_word()) {
    case nul:
      check_surjections();
      return;
   case srt:
      get_item(get_sort,0,0);
      break;
    case fcn:
      get_item(get_func,0,0);
      break;
    case fcn0:
      get_item(get_const,0,0);
      break;
    case vcb:
      get_item(get_vocab,0,0);
      break;
    case cls:
      get_item(get_input_cl,0,0);
      break;
    case prm:
      get_item(get_set,0,0);
      break;
    case inc:
      get_incl_file();
      break;
    }
} /* read_job */



/*------------------------------------------------------------------------*/
void check_surjections(void)
{
  functor f;
  dynstring ds;
  char s[1000];

  for (f=FN; f; f=f->nextfunct)
    if (!f->surj_set) {
      if (f->adicity == 1 && f->surjective && !f->injective) {
	f->surj_set = true;
	dynstring_initialise(&ds,"");
	if (setting.function_spec_mode == math)
	  sprintf(s,"function @surj_skolem_%s: %s -> %s {injective hidden} ",
		  f->fname,f->val_type->sname,f->arg_type[0]->sname);
	else
	  sprintf(s,"vocabulary function @surj_skolem_%s(%s): %s {injective hidden} ",
		  f->fname,f->val_type->sname,f->arg_type[0]->sname);
	dynstring_strcpyc(&ds,s);
	sprintf(s,"clause { ALL x@@(%s(@surj_skolem_%s(x@@)) = x@@). }",
		f->fname,f->fname);
	dynstring_strcatc(&ds,s);
	finder_input(&ds);
	dynstring_free(&ds);
      }
      /* MORE TO COME IN HERE */
    }
}


/*------------------------------------------------------------------------*/
void get_incl_file(void)

     /*Called by read_job().
      *Get_incl causes a file to be read as though it were
      *included at the current point. Note that it must end
      *with the word `end'. The file name is enclosed in 
      *(round) parentheses.
      *
      *At present, the included file must be external.
      */
     
     
{
  FILE *oldsource, *infil;
  ifile oldisource;
  dynstring include_filename;
  dynstring oldjobline;
  int oldinputpoint, i;

  dynstring_initialise(&oldjobline, "");
  dynstring_initialise(&include_filename, "");
  
  get_word(&include_filename);
  if (include_filename.c_str[0] != '(')
    inerror("Included filename must be in parentheses","",0);
  
  i = 0;
  dynstring_ensure_space(&include_filename, 0);
  while ((include_filename.c_str[i] = get_a_char()) != ')') {
    i++;
    dynstring_ensure_space(&include_filename, i);
  }
  include_filename.c_str[i] = '\0';
  
  oldsource = source;
  oldisource = isource;
  isource = 0;
  dynstring_strcpy(&oldjobline, &job_line);
  oldinputpoint = input_point;
  if (!(infil = fopen(include_filename.c_str,"r")))
    inerror("Cannot open file",include_filename.c_str,0);
  read_job(infil,0);
  fclose(source);
  source = oldsource;
  isource = oldisource;
  dynstring_strcpy(&job_line, &oldjobline);
  input_point = oldinputpoint;

  dynstring_free(&oldjobline);
  dynstring_free(&include_filename);

} /* get_incl_file() */


/*
*	The final trivial function in this section allows a calling
*	program to add FINDER input one line at a time.
*/

void finder_input(const dynstring *s)
{
  dynstring temp;

  dynstring_initialise(&temp, "");
  cmd_line_input = true;
  input_point = 0;
  dynstring_strcpy(&temp,&command_line);
  dynstring_strcpy(&command_line,s);
  dynstring_strcatc(&command_line," end.");
  read_job(0,0);
  dynstring_strcpy(&command_line,&temp);
}



/*------------------------------------------------------------------------*/
int get_int(void)

     /*Called by get_set(), get_item(), sort_spec(), func_spec(), 
      *and got_function().
      *Gets an expected unsigned integer charater from the input line.
      *Ignores leading colons, as these are optional.
      *Exits in error if one is not found.*/

{
  dynstring s;
  int rc;

  dynstring_initialise(&s, "");
  do
    get_word(&s);
  while (s.c_str[0] == ':');
  if (!sscanf(s.c_str,"%d",&rc))
    inerror(s.c_str,"found where an integer was expected",0);
  dynstring_free(&s);
  return rc;
} /*get_int()*/

/*------------------------------------------------------------------------*/
float get_float(void)

     /*Not used at present. Could be used to get priorities etc.
      *Gets an expected float from the input line.
      *Exits in error if one is not found.*/
{
  dynstring s;
  float rc;

  dynstring_initialise(&s, "");
  do
    get_word(&s);
  while (s.c_str[0] == ':');
  if (!sscanf(s.c_str,"%f",&rc))
    inerror(s.c_str,"found where a number was expected",0);
  dynstring_free(&s);
  return rc;
} /*get_float()*/


/*------------------------------------------------------------------------*/
void get_item(void (*item)(), kind k, functor f)

     /*Called by read_job(), sort_spec(), get_func(), get_cons(), 
      *and get_set().

      *Calls the appropriate function (*item)() to get either
      *one item of the given species (e.g. one clause) or a set of them.
      *A set, naturally, is specified by listing its members within a 
      *pair of braces.
      *
      *Some items like clauses have an optional integer parameter given
      *in parentheses after the key word, so get_item must search for 
      *this first.
      *
      *Functions called by get_item have all four of the parameters it 
      *issues, but generally don't use most of them.  Although the above
      *function read_job does not make use of the parameters k and f, 
      *functions (*item)() may do so when they themselves call get_item 
      *for their sub-items.
      *All calls to (*item)() are pased the first word of the item.*/

{
  dynstring s;
  int arg = 0;

  dynstring_initialise(&s, "");

  get_word(&s);
  if (s.c_str[0] == '(') {  /* initial arg of item is an int in parentheses */
    get_word(&s);
    if (!strcmp(s.c_str,"low")) arg = LOW_WEIGHT;
    else if (!strcmp(s.c_str,"med")) arg = MED_WEIGHT;
    else if (!strcmp(s.c_str,"high")) arg = HIGH_WEIGHT;
    else if (!sscanf(s.c_str,"%d",&arg))
      inerror(s.c_str,"found where an integer was expected",0);
    get_word(&s); /*get closing parentheses */
    if (s.c_str[0] != ')') inerror("Missing right parenthesis","",0);
    get_word(&s); /*get string following integer argument */
  }

  if (s.c_str[0] == '{') {  /* list of items in braces */
    get_word(&s);
    while (s.c_str[0] != '}') {
      if (s.c_str[0] != '.')
	(*item)(k,f,&s,arg);
      get_word(&s);  /* get next item */
    }
  }
  else if (s.c_str[0] != '.') (*item)(k,f,&s,arg);   /* single item */

  dynstring_free(&s);
}



/*------------------------------------------------------------------------*/
void get_sort(kind k, functor f, dynstring *name, int x)

     /*Called by get_item().
      *Creates a new sort if "name" does'nt already exist.
      *Gets a (possibly) revised specification for that sort.*/
{
  get_item(sort_spec, newkind(name->c_str), 0);
} /* get_sort() */



/*------------------------------------------------------------------------*/
void sort_spec(kind k, functor f, dynstring *spec, int x)

     /* Called by get_item().
      *	Revises the specifications (if any) for the kind k.*/

{
  int i;

  if (!strcmp(spec->c_str,"cardinality") || !strcmp(spec->c_str,"card")) {
    get_word(spec);
    switch (spec->c_str[0]) {
    case '=':
      i = get_int();
      if (!k->canonical)
	k->sizmax = k->sizmin = i;
      return;
    case '<':
      i = get_int();
      if (!k->canonical) {
	k->sizmax = i-1;
	if (k->sizmax < k->sizmin)
	  k->sizmin = k->sizmax;
      }
      return;
    case '>':
      i = get_int();
      if (!k->canonical) {
	k->sizmin = i + 1;
	if (k->sizmax < k->sizmin)
	  k->sizmax = k->sizmin;
      }
      return;
    }
    inerror("","Statement of cardinality needs =, < or >",0);		
  }
  else if (!strcmp(spec->c_str,"enum")) {
    if (k==Bool || k==Integer) {
      inerror("Attempt to set enumeration of builtin sort",k->sname,0);
      return;
    }
    de_enumerate(k);
    for (;;) {
      get_word(spec);
      if (spec->c_str[0] == ':')
	get_word(spec);
      if (isdigit(spec->c_str[0])) {
	for (i=1; spec->c_str[i]; i++)
	  if (!isdigit(spec->c_str[i])) {
	    inerror(spec->c_str,"starts with a digit but is not a number",0);
	    return;
	  }
      }
      if (isdigit(spec->c_str[0]) ||
	  !strcmp(spec->c_str,"TRUE") || !strcmp(spec->c_str,"FALSE")) {
	inerror(spec->c_str,"is a reserved word and not allowed in an enum",0);
	return;
      }
      canon(k,spec->c_str);
      get_word(spec);
      if (spec->c_str[0] == '.')
	return;
      if (spec->c_str[0] != ',')
	inerror(spec->c_str,"found where comma expected",0);
    }
  }
  else if (!strcmp(spec->c_str,"CrashDummy")) {
    fprintf(stderr,"Dummy crash\n");
    {kind x; KIND y;
      for (x=0; x<&y; x++)
	x->siz = *(x->sum);
    }
  }
  else
    inerror(spec->c_str,
	    "found where key word \"cardinality\" or \"enum\" expected",0);
} /* sort_spec() */



/*------------------------------------------------------------------------*/
void de_enumerate(kind k)

     /*Called by sort_spec().
      *Removes the existing enumeration (if any) of kind k. */
{
  free_canon(k->canonical);
  k->canonical = 0;

} /* de_enumerate() */


/*------------------------------------------------------------------------*/
void free_canon(name_list lp)

     /*Called by de_emumerate().
      *Recusively frees pointers to the cannical names in the 
      *enumeration list. */
{
  if (!lp) return;
  free_canon(lp->nextname);
  free(lp->nextname);
  lp->nextname = 0;

} /* free_canon() */



/*------------------------------------------------------------------------*/
void get_func(kind k, functor f, dynstring *name, int x)
     
     /*Called by get_item(), and get_cons().
      *Scans for the argument and value types of the (new) functor. 
      *If the functor all ready exists then this is redefined.
      *Since newfuntor() is called and sets the print mode to null
      *(which it needs to do because it usually sets up variables, etc) 
      *so this function resets it to the default full.
      *Then calls func_spec() to scan any function specifications.
 
      *Parameters k, f and x are dummies for get_func.  The string s
      *holds the name of the function. */
     
{
  switch(setting.function_spec_mode) {
  case math:
    get_func_math(k,f,name,x);
    break;
  case comp:
    get_func_comp(k,f,name,x);
    break;
  default:
    inerror("Bad function specification mode",0,0);
  }
}


void get_func_math(kind k, functor f, dynstring *name, int x)
     
/* Version of get_func where input mode is math */
     
{
  int i;
  int arity=0;
  boolean got_arg_types=false;
  boolean got_val_type=false;
  dynstring type_str, punctuation;
  dynstring fcn_types[ARGMAX+1];

  dynstring_initialise(&type_str, "");
  dynstring_initialise(&punctuation, "");
  
  for (i = 0; i <= ARGMAX; i++) dynstring_initialise(&fcn_types[i], "");
  
  get_word(&type_str);
  
  if (type_str.c_str[0] == ':') get_word(&type_str); /* skip ':' */

  /* get the types of the functor's arguments, and then its value type */
  do {
#ifdef SCOTT
    if (arity > ARGMAX){
      inerror("Adicity too great for function",name->c_str,0);
      over_argmax = true; 
      return;
    }
#else
    if (arity > ARGMAX) inerror("Adicity too great for function",name->c_str,0);
#endif
    dynstring_strcpy(&fcn_types[arity++],&type_str); /*record the next type*/
    if (got_arg_types) {
      got_val_type = true;
      arity--; /* the functions values is stored as the last type */
    }
    else {
      get_word(&punctuation);

      if (!strchr(".,-",punctuation.c_str[0]))
	inerror(type_str.c_str,"found where punctuation (, -> or .) expected",0);

      switch(punctuation.c_str[0]) {
      case '.':
	arity--;
	got_val_type = true;
	break;
      case ',': 
	if (got_arg_types) inerror(name->c_str,"can only have ONE value type",0);
	get_word(&type_str); /* get next argument type */
	break;
      case '-':
	if (punctuation.c_str[1] != '>')
	  inerror(punctuation.c_str,"found where \"->\" expected",0);
	got_arg_types = true;
	get_word(&type_str); /* get value type */
	break;
      }
    }
  }
  while (!got_val_type);

  f = newfunctor(name->c_str, Func, arity, true);
  f->val_type = set_up_type(&fcn_types[arity]);
  for (i=0; i<arity; i++)
    f->arg_type[i] = set_up_type(&fcn_types[i]);
  f->pmode = full;
  get_item(func_spec, 0, f);
  func_sanity_check(f);

  dynstring_free(&type_str);
  dynstring_free(&punctuation);
  
  for (i = 0; i <= ARGMAX; i++) dynstring_free(&fcn_types[i]);
  
} /* get_func */



/*------------------------------------------------------------------------*/
void get_pred(kind k, functor f, dynstring *name, int x)
/*
Like get_func but with the knowledge that the value type is bool.
*/

{
  switch(setting.function_spec_mode) {
  case math:
    get_pred_math(k,f,name,x);
    break;
  case comp:
    get_pred_comp(k,f,name,x);
    break;
  default:
    inerror("Bad function specification mode",0,0);
  }
}


void get_pred_math(kind k, functor f, dynstring *name, int x)
{
  int i;
  int arity=0;
  dynstring type_str, punctuation;
  dynstring fcn_types[ARGMAX+1];

  dynstring_initialise(&type_str, "");
  dynstring_initialise(&punctuation, "");  
  for (i = 0; i <= ARGMAX; i++) dynstring_initialise(&fcn_types[i], "");
  
  get_word(&type_str);
  
  if (type_str.c_str[0] == ':') get_word(&type_str); /* skip ':' */

  /* get the types of the functor's arguments, and then its value type */
  for(;;) {
    if (arity > ARGMAX) {
      inerror("Adicity too great for function",name->c_str,0);
#ifdef SCOTT
      over_argmax = true; 
#endif
      return;
    }

    dynstring_strcpy(&fcn_types[arity++],&type_str); /*record the next type*/
    get_word(&punctuation);
    if (punctuation.c_str[0] == '.')
      break;
    else if (punctuation.c_str[0] == ',')
      get_word(&type_str); /* get next argument type */
    else
      inerror(type_str.c_str,"found where punctuation (, or .) expected",0);
    }

  f = newfunctor(name->c_str, Func, arity, true);
  f->val_type = Bool;
  for (i=0; i<arity; i++)
    f->arg_type[i] = set_up_type(&fcn_types[i]);
  f->pmode = full;
  get_item(func_spec, 0, f);
  func_sanity_check(f);

  dynstring_free(&type_str);
  dynstring_free(&punctuation);  
  for (i = 0; i <= ARGMAX; i++) dynstring_free(&fcn_types[i]);
  
} /* get_pred */



/*------------------------------------------------------------------------*/
void get_const(kind k, functor f, dynstring *name, int x)
     
     /*Called by get_item().
      *Calls get_func() for 0-adic function symbol. */

{
  dynstring const_type;

  dynstring_initialise(&const_type, "");

  f = newfunctor(name->c_str, Func, 0, true);

  do	get_word(&const_type);
  while (const_type.c_str[0] == ':');

  f->val_type = set_up_type(&const_type);
  f->pmode = full;
  get_item(func_spec, 0, f);
  func_sanity_check(f);

  dynstring_free(&const_type);

} /* get_const() */




/*------------------------------------------------------------------------*/
void get_func_comp(kind k, functor f, dynstring *name, int x)

/*
Variant of get_func using CS notation f(X): Y.
*/

{
  int i;
  int arity=0;
  dynstring type_str, punctuation;
  dynstring fcn_types[ARGMAX+1];

  dynstring_initialise(&type_str, "");
  dynstring_initialise(&punctuation, "");  
  for (i = 0; i <= ARGMAX; i++) dynstring_initialise(&fcn_types[i], "");

  get_word(&punctuation);
  if (punctuation.c_str[0] == '(') {
    // Get the argument types
    do {
      get_word(&type_str);
      if (arity > ARGMAX)
	inerror("Adicity too great for function",name->c_str,0);
      if (!arity && type_str.c_str[0] == ')')
	dynstring_strcpy(&punctuation,&type_str);
      else {
	dynstring_strcpy(&fcn_types[arity++],&type_str); /*record the next type*/
	get_word(&punctuation);
      }
      if (!strchr(",)",punctuation.c_str[0]))
	inerror(type_str.c_str,
		"found in function definition where \",\" or \")\" expected",0);
    }
    while (punctuation.c_str[0] == ',');
  }
  else
    inerror(punctuation.c_str,
	    "found in function definition where \"(\" expected\n",0);
  get_word(&punctuation);
  if (punctuation.c_str[0] != ':')
    inerror(punctuation.c_str,
	    "found in function definition where \":\" expected",0);
  get_word(&type_str);

  f = newfunctor(name->c_str, Func, arity, true);
  f->val_type = set_up_type(&type_str);
  for (i=0; i<arity; i++)
    f->arg_type[i] = set_up_type(&fcn_types[i]);
  f->pmode = full;
  get_item(func_spec, 0, f);
  func_sanity_check(f);

  dynstring_free(&type_str);
  dynstring_free(&punctuation);  
  for (i = 0; i <= ARGMAX; i++) dynstring_free(&fcn_types[i]);
  
} /* get_func_comp */



/*------------------------------------------------------------------------*/
void get_pred_comp(kind k, functor f, dynstring *name, int x)
     
/*
  As get_func_comp with understanding that the value type is Bool
*/

{
  int i;
  int arity=0;
  dynstring type_str, punctuation;
  dynstring fcn_types[ARGMAX+1];

  dynstring_initialise(&type_str, "");
  dynstring_initialise(&punctuation, "");  
  for (i = 0; i <= ARGMAX; i++) dynstring_initialise(&fcn_types[i], "");
  
  get_word(&punctuation);
  if (punctuation.c_str[0] == '(') {
    // Get the argument types
    do {
      get_word(&type_str);
      if (arity > ARGMAX)
	inerror("Adicity too great for function",name->c_str,0);
      if (!arity && type_str.c_str[0] == ')')
	dynstring_strcpy(&punctuation,&type_str);
      else {
	dynstring_strcpy(&fcn_types[arity++],&type_str); /*record the next type*/
	get_word(&punctuation);
      }
      if (!strchr(",)",punctuation.c_str[0]))
	inerror(punctuation.c_str,
		"found in predicate definition where \",\" or \")\" expected",0);
    }
    while (punctuation.c_str[0] == ',');
  }
  else
    inerror(type_str.c_str,
	    "found where \"(\" expected in predicate definition",0);

  f = newfunctor(name->c_str, Func, arity, true);
  f->val_type = Bool;
  for (i=0; i<arity; i++)
    f->arg_type[i] = set_up_type(&fcn_types[i]);
  f->pmode = full;
  get_item(func_spec, 0, f);
  func_sanity_check(f);

  dynstring_free(&type_str);
  dynstring_free(&punctuation);  
  for (i = 0; i <= ARGMAX; i++) dynstring_free(&fcn_types[i]);

} /* get_pred_comp */



/*------------------------------------------------------------------------*/
void get_vocab(kind k, functor f, dynstring *species, int x)
     
     /*Called by get_item().
      *Calls get_func() for species of function symbol(s). */

{
  if (!strncmp(species->c_str,"name",4)) {
    get_item(get_const,0,0);
  }
  else if (!strncmp(species->c_str,"predicate",9)) {
    get_item(get_pred,0,0);
  }
  else if (!strncmp(species->c_str,"function",8)) {
    get_item(get_func,0,0);
  }
  else
    inerror(species->c_str,
	    "found where \"name\", \"predicate\" or \"function\" expected",0);

} /* get_vocab() */




/*------------------------------------------------------------------------*/
kind set_up_type(dynstring *type)

     /*Called by get_func() and get_const().
      *Returns a pointer to the kind coresponding to the given "type".*/
{
  int i;
  
  for (i = 0; sort[i].sname; i++) 
    if (!strcmp(sort[i].sname,type->c_str)) return(&sort[i]);
  inerror("Non-existent sort",type->c_str,0);
  return 0;
} /* set_up_type() */




/*------------------------------------------------------------------------*/
void func_spec(kind k, functor f, dynstring *key_word, int x)

     /* Called by get_func() and get_const().
      *Sets up the various properties of the function that can be 
      *specified in the section following its declaration. 
      *(The paramters x and f are just dummies). */
     
{
  dynstring specs;
  int i;

  dynstring_initialise(&specs, "");
  
  if (!strcmp(key_word->c_str,"total")) f->total = true;
  else if (!strcmp(key_word->c_str,"partial")) f->total = false;
  else if (!strcmp(key_word->c_str,"injective") ||
	    !strcmp(key_word->c_str,"all_different")) f->injective = true;
  else if (!strcmp(key_word->c_str,"left_injective") || 
	    !strcmp(key_word->c_str,"left_all_different") || 
	    !strcmp(key_word->c_str,"column_injective") || 
	    !strcmp(key_word->c_str,"column_all_different"))
    f->leftin = (f->adicity == 2);
  else if (!strcmp(key_word->c_str,"right_injective") || 
	    !strcmp(key_word->c_str,"right_all_different") || 
	    !strcmp(key_word->c_str,"row_injective") || 
	    !strcmp(key_word->c_str,"row_all_different"))
    f->rightin = (f->adicity == 2);
  else if (!strcmp(key_word->c_str,"surjective")) f->surjective = true;
  else if (!strcmp(key_word->c_str,"left_surjective") || 
	    !strcmp(key_word->c_str,"column_surjective"))
    f->leftsur = (f->adicity == 2);
  else if (!strcmp(key_word->c_str,"right_surjective") || 
	    !strcmp(key_word->c_str,"row_surjective"))
    f->rightsur = (f->adicity == 2);
  else if (!strcmp(key_word->c_str,"bijective") ||
	    !strcmp(key_word->c_str,"one_one"))
    f->surjective = f->injective = true;
  else if (!strcmp(key_word->c_str,"left_bijective") || 
	    !strcmp(key_word->c_str,"left_one_one") || 
	    !strcmp(key_word->c_str,"column_bijective") || 
	    !strcmp(key_word->c_str,"column_one_one"))
    f->leftsur = f->leftin = (f->adicity == 2);
  else if (!strcmp(key_word->c_str,"right_bijective") || 
	    !strcmp(key_word->c_str,"right_one_one") || 
	    !strcmp(key_word->c_str,"row_bijective") || 
	    !strcmp(key_word->c_str,"row_one_one"))
    f->rightsur = f->rightin = (f->adicity == 2);
  else if (!strcmp(key_word->c_str,"commutative")) {
    f->comm = true;
    for (i=1; i<f->adicity; i++)
      if (f->arg_type[i] != f->arg_type[0])
	f->comm = false;
  }
  else if (!strcmp(key_word->c_str,"cut")) f->cut_it = true;
  else if (!strcmp(key_word->c_str,"no_cut"))
    f->cut_it = false;
  else if (!strcmp(key_word->c_str,"priority")) f->nop = false;
  else if (!strcmp(key_word->c_str,"no_priority")) f->nop = true;
  else if (!strcmp(key_word->c_str,"labels")) f->labels = true;
  else if (!strcmp(key_word->c_str,"no_labels")) f->labels = false;
  else if (!strcmp(key_word->c_str,"cgv")) {
    f->cgv = true;
    setting.cgv_enabled = true;
  }
  else if (!strcmp(key_word->c_str,"no_cgv")) f->cgv = false;
  else if (!strcmp(key_word->c_str,"configure")) {
    f->configurable = true;
    setting.configmode = true;
  }
  else if (!strcmp(key_word->c_str,"no_configure")) f->configurable = false;
  else if (!strcmp(key_word->c_str,"print")) {
    get_word(&specs);
    if (specs.c_str[0] == ':') get_word(&specs);
    if (!strcmp(specs.c_str,"none")) {
      f->pmode = null;
    }
    else if (!strcmp(specs.c_str,"brief")) f->pmode = brief;
    else if (!strcmp(specs.c_str,"full")) f->pmode = full;
    else inerror(specs.c_str,"found where print mode expected",0);
    return;
  }
  else if (!strcmp(key_word->c_str,"hidden")) {
    f->pmode = null;
    f->cut_it = true;
  }
  else if (!strcmp(key_word->c_str,"scope")) {
    f->scope = get_int();
    if (f->scope > 10) f->scope = 10;
    return;
  }
  else inerror(key_word->c_str,"found where function specification expected",0);

  dynstring_free(&specs);

} /* func_spec() */


/*------------------------------------------------------------------------*/
boolean add_clause(dynstring *clause_string)

     /*Called by model() and find_theories().
      *Parse clause_string as tho' it was next on the input line. 
      *Returns TRUE iff clause is non-trivial*/

{
  boolean result = false;
  dynstring lead_word;

  dynstring_initialise(&lead_word, "");

  dynstring_strcpy(&job_line,clause_string);
  input_point = 0;

  get_word(&lead_word);
  if (get_cl(&lead_word,0)) {
    result = true;
  } else {
    inwarning("\nPassed clause is trivially satisfiable.","");
  }

  dynstring_free(&lead_word);
  
  return result;

}/* add_clause() */


void add_declaration(dynstring *declaration)
  /*----------------------------------------------------------------------*\
    Called by SCOTT fcn declare_otter_symbols().
    Parse "declaration" as tho' it was next on the input line. 
  \*----------------------------------------------------------------------*/
{
  dynstring_strcpy(&job_line,declaration);
  input_point = 0;
  source = NULL; // not reading from a file

  switch(get_key_word()) {
  case nul:
    return;
  case srt:
    get_item(get_sort,0,0);
    break;
  case fcn:
    get_item(get_func,0,0);
    break;
  case fcn0:
    get_item(get_const,0,0);
    break;
  case vcb:
    get_item(get_vocab,0,0);
    break;
  case cls:
    get_item(get_input_cl,0,0);
    break;
  case prm:
    get_item(get_set,0,0);
    break;
  case inc:
    get_incl_file();
    break;
  }

}/*--add_declaration()----------------------------------------------------*/


/*------------------------------------------------------------------------*/
void get_input_cl(kind k, functor f, dynstring *first_word, int softweight)

     /*Called by get_item(). 
      *Get the clause off the input line and, provided it's non-trivial,
      *add it to the current theory.*/
     
{
  if (first_word->c_str[0] != '.' &&
      !strchr(job_line.c_str+input_point,'.') && 
      !strchr(job_line.c_str+input_point,'}'))
    input_to_terminator(".}");

  if (get_cl(first_word,softweight)) {
    if (normal_input)
      add_to_theory();
    else
      re_clausify();
  }
  else 
    inwarning("Input clause is trivially satisfiable.","");
}


/*------------------------------------------------------------------------*/
boolean get_cl(dynstring *lead_word, int softweight)

     /*Called by get_input_cl() and add_clause().
      *Parse the next clause on the input line into latest_clause.
      *Returns TRUE iff the clause is non-trivial.*/ 

{
  wff_stack *ins = 0;
  wff tempwff = 0 ;
  boolean parsed_premises;

  latest_clause.softweight = softweight;
  latest_clause.pid = next_pid++;
  begin_parse();
  parsed_premises = true;
  if (strstr(job_line.c_str+input_point,"->")) {
    if (setting.single_literal)
      inerror("Clause-forming arrow (->) is not allowed in FOF form","",0);
    else
      parsed_premises = false;
  }

  do {
    if (strcmp(lead_word->c_str,"->")) {
      tempwff = parse(lead_word,0);
      if (parsed_premises)
	ins = &(latest_clause.conc);
      else
	ins = &(latest_clause.prem);
    }

    if (setting.single_literal && strcmp(lead_word->c_str,"."))
      inerror(lead_word->c_str,"found where \".\" expected",0);

    if (!strcmp(lead_word->c_str,"->")) {
      if (parsed_premises)
	inerror("Apparently nested arrows in clause","",0);
      else
	parsed_premises = true;
    }
    else if (!strcmp(lead_word->c_str,",") && parsed_premises)
      inwarning("",
		"Consequents should be separated by semicolon, not comma");
    else if (!strcmp(lead_word->c_str,";") && !parsed_premises)
      inwarning("",
		"Antecedents should be separated by comma, not semicolon");
    else if (!strchr(",;.",lead_word->c_str[0]))
      inerror(lead_word->c_str,"found in clause where punctuation expected",0);

    if (tempwff) {
      while (*ins) ins = &((*ins)->nxt);
      *ins = (wff_stack) malloc(sizeof(WFF_STACK));
      (*ins)->w = tempwff;
      (*ins)->nxt = 0;
    }

    if (strcmp(lead_word->c_str,"."))
      get_word(lead_word); /*get next predicate name*/
  }
  while (strcmp(lead_word->c_str,"."));   /* all clauses terminated with "." */

  typecheck();
  end_parse(&latest_clause);

  return true;

} /* get_cl() */


void re_clausify(void)
// latest_clause comes from found with abnormal input, so it should be
// a single clause made into a formula by the print routines in
// foutput.c so it may contain quantifiers etc. Turn it back into a
// pure clause. This should be quasi-instant, in realistic cases.
{
  CLAUSE all_latest_clauses[CL_MAX];   /* Clausified version */
  int alc_size = 1;
  int clause_id;

  /* initialise, in case these have junk values */
  for (clause_id = 0; clause_id < CL_MAX; clause_id++) {
    all_latest_clauses[clause_id].prem = 0;
    all_latest_clauses[clause_id].conc = 0;
    all_latest_clauses[clause_id].varlist = 0;
    all_latest_clauses[clause_id].testlist = 0;
  }
  alc_size = clausify(&latest_clause,all_latest_clauses);
  if (alc_size > 1)
    inerror("Multiple clauses passed to tester with abnormal input","",0);
  free_clause_parts(&latest_clause);
  latest_clause.prem = all_latest_clauses[0].prem;
  latest_clause.conc = all_latest_clauses[0].conc;
  latest_clause.varlist = all_latest_clauses[0].varlist;
  latest_clause.testlist = all_latest_clauses[0].testlist;
}



/*------------------------------------------------------------------------*/
boolean isint(char *s)
     
     /*Called by one line in get_set(). This could be dodgy. */
     
{
  do if (!isdigit(*s)) return false;
  while (*(++s));
  return true;
}/* isint() */


/*------------------------------------------------------------------------*/
void get_set(kind k, functor f, dynstring *set, int x)
     
     /*Called by get_item().
      *Extracts setting information.*/
     
{
  if (!strcmp(set->c_str,"solutions")) setting.max_models = get_int();
  else if (!strcmp(set->c_str,"backtracks")) 
    setting.max_backtracks = get_int();
  else if (!strcmp(set->c_str,"insertions")) 
    setting.max_invalues = get_int();
  else if (!strcmp(set->c_str,"flips")) 
    setting.max_flips = get_int();
  else if (!strcmp(set->c_str,"stack")) {
    get_word(set);
    if (set->c_str[0] == ':') get_word(set);
    if (!strcmp(set->c_str,"small")) setting.batches = S_SMALL;
    else if (!strcmp(set->c_str,"medium")) setting.batches = S_MEDIUM;
    else if (!strcmp(set->c_str,"large") || !strcmp(set->c_str,"LARGE"))
      setting.batches = S_LARGE;
    else if (!strcmp(set->c_str,"maximal")) setting.batches = S_MAXIMAL;
    else if (isint(set->c_str)) sscanf(set->c_str,"%d",&(setting.batches));
    else inerror(set->c_str,"found where stack size expected",0);
    return;
  }
  else if (!strcmp(set->c_str,"stack_margin")) 
    setting.extra_batches = get_int();
  else if (!strcmp(set->c_str,"verbosity")) get_item(verb_spec, 0, 0);
  else if (!strcmp(set->c_str,"ref_length")) setting.maxref = get_int();
  else if (!strcmp(set->c_str,"pre_test")) setting.pretest_limit = get_int();
  else if (!strcmp(set->c_str,"test_ref_length")) 
    setting.test_maxref = get_int();
  else if (!strcmp(set->c_str,"time_limit")) setting.maxtime = get_int();
  else if (!strcmp(set->c_str,"cv_limit")) setting.max_cellvals = get_int();
  else if (!strcmp(set->c_str,"max_bound")) setting.target_bound = get_int();
  else if (!strcmp(set->c_str,"dump_rate")) {
    setting.dump_rate = get_int();
    if (setting.dump_rate) {
      setting.pm.verbosity[pjob] = null;
      setting.pm.verbosity[pmodels] = null;
      setting.pm.verbosity[pstats] = full;
    }
  }

  else if (!strcmp(set->c_str,"no_process")) setting.parse_only = true;
  else if (!strcmp(set->c_str,"process")) setting.parse_only = false;
  
  else if (!strcmp(set->c_str,"no_search")) setting.clausify_only = true;
  else if (!strcmp(set->c_str,"search")) setting.clausify_only = false;
  
  else if (!strcmp(set->c_str,"primary")) setting.Nosecs = true;
  else if (!strcmp(set->c_str,"secondary")) setting.Nosecs = false;
  
  else if (!strcmp(set->c_str,"noisy")) setting.noisy = true;
  else if (!strcmp(set->c_str,"quiet")) setting.noisy = false;
  
  else if (!strcmp(set->c_str,"one_cardinality")) setting.sortcut = true;
  else if (!strcmp(set->c_str,"all_cardinalities")) setting.sortcut = false;
  
  else if (!strcmp(set->c_str,"time_stamp")) setting.timestamps = true;
  else if (!strcmp(set->c_str,"no_time_stamp")) setting.timestamps = false;
  
  else if (!strcmp(set->c_str,"profile")) setting.profile = true;
  else if (!strcmp(set->c_str,"no_profile")) setting.profile = false;

  else if (!strcmp(set->c_str,"dump_depth")) setting.dumpdepth = true;
  else if (!strcmp(set->c_str,"no_dump_depth")) setting.dumpdepth = false;

  else if (!strcmp(set->c_str,"config_heuristics")) setting.configheuristic = true;
  else if (!strcmp(set->c_str,"no_config_heuristics")) setting.configheuristic = false;

  else if (!strcmp(set->c_str,"awnty")) setting.awnty_enabled = true;
  else if (!strcmp(set->c_str,"no_awnty")) setting.awnty_enabled = false;

  else if (!strcmp(set->c_str,"report_breaks")) setting.report_breaks = true;
  else if (!strcmp(set->c_str,"silent_breaks")) setting.report_breaks = false;
  
  else if (!strcmp(set->c_str,"print_headers")) setting.print_headers = true;
  else if (!strcmp(set->c_str,"no_print_headers")) setting.print_headers = false;
  
  else if (!strcmp(set->c_str,"print_unsat")) setting.print_unsat = true;
  else if (!strcmp(set->c_str,"no_print_unsat")) setting.print_unsat = false;
  
  else if (!strcmp(set->c_str,"local_search")) {
    setting.search_is_local = true;
    if (setting.max_flips == 0) setting.max_flips = MAX_FLIPS_DEFAULT;
  }
  else if (!strcmp(set->c_str,"global_search")) {
    setting.search_is_local = false;
    setting.max_flips = 0;
  }
  
  else if (!strcmp(set->c_str,"tops_off")) setting.topsoff = true;
  else if (!strcmp(set->c_str,"tops_on")) setting.topsoff = false;
  
  else if (!strcmp(set->c_str,"reverse_test")) setting.invert_change = true;
  else if (!strcmp(set->c_str,"obverse_test")) setting.invert_change = false;
  
  else if (!strcmp(set->c_str,"test_by_vars")) setting.vars_over_wt = true;
  else if (!strcmp(set->c_str,"test_by_weight")) setting.vars_over_wt = false;

  else if (!strcmp(set->c_str,"random")) {
    setting.random = true;
    srand48((long) get_int());
  }
  else if (!strcmp(set->c_str,"non_random"))
    setting.random = false;
  
  else if (!strcmp(set->c_str,"pretty")) setting.pm.pretty = true;
  else if (!strcmp(set->c_str,"ugly")) {
    setting.pm.pretty = false;
    setting.pm.number_only = false;
    setting.pm.verbosity[pjob] = brief;
    setting.pm.verbosity[pmodels] = brief;
    setting.pm.verbosity[pstats] = brief;
    setting.pm.verbosity[pspace] = null;
  }

  else if (!strcmp(set->c_str,"print_number_only")) {
    setting.pm.number_only = true;
    setting.pm.verbosity[pjob] = null;
    setting.pm.verbosity[pmodels] = null;
    setting.pm.verbosity[pstats] = brief;
    setting.pm.verbosity[pspace] = null;
  }

  else if (!strcmp(set->c_str,"anytime")) setting.pm.incremental = true;
  else if (!strcmp(set->c_str,"onetime")) setting.pm.incremental = false;

  else if (!strcmp(set->c_str,"debug"))
    get_item(event_subscription, 0, 0);

  //else inerror("Unrecognised setting option",set->c_str,0);

} /* get_set() */



/*------------------------------------------------------------------------*/
void verb_spec(kind k, functor f, dynstring *s, int x)

     /*Called by get_set() only.*/
{
  if (!strcmp(s->c_str,"models"))
    verb_spec_spec(pmodels);
  else if (!strcmp(s->c_str,"job"))
    verb_spec_spec(pjob);
  else if (!strcmp(s->c_str,"stats"))
    verb_spec_spec(pstats);
  else if (!strcmp(s->c_str,"space"))
    verb_spec_spec(pspace);
  else
    inerror("Unrecognised verbosity option",s->c_str,0);

} /*verb_spec()*/



/*------------------------------------------------------------------------*/
void verb_spec_spec(print_mode printing)

/*Called by verb_spec() only. */

{
  dynstring s;

  dynstring_initialise(&s, "");

  get_word(&s);
  if (s.c_str[0] == ':') get_word(&s);

  if (!strcmp(s.c_str,"brief"))
    setting.pm.verbosity[printing] = brief;
  else if (!strcmp(s.c_str,"full")) {
    setting.pm.verbosity[printing] = full;
    setting.pm.number_only = false;
  }
  else if (!strcmp(s.c_str,"none"))
    setting.pm.verbosity[printing] = null;
  else {
    inerror(s.c_str, "found where level (none, brief or full) expected",0);
  }
  dynstring_free(&s);

} /* verb_spec_spec() */


/*------------------------------------------------------------------------*/
void event_subscription(kind k, functor f, dynstring *s, int x)

     /*Called by get_set() only.*/
{
  debug_event e;

  e = event_called(s->c_str);
  if (!e)
    inerror(s->c_str, "found where debug event name expected",0); 
  event_subs_spec(e);
} /*event_subscription*/


/*------------------------------------------------------------------------*/
void event_subs_spec(debug_event e)

/*Called by event_subscription() only. */

{
  dynstring s;

  dynstring_initialise(&s, "");

  get_word(&s);
  if (s.c_str[0] == ':') get_word(&s);

  if (!strcmp(s.c_str,"collect"))
    e->subscribed = collect_only;
  else if (!strcmp(s.c_str,"block"))
    e->subscribed = blocking_send;
  else if (!strcmp(s.c_str,"noblock"))
    e->subscribed = non_blocking_send;
  else if (!strcmp(s.c_str,"none"))
    e->subscribed = unsubscribed;
  else {
    inerror(s.c_str, "found where level (none, collect, noblock, block) expected",0);
  }
  dynstring_free(&s);

} /* event_subs_spec() */



/*------------------------------------------------------------------------*/
boolean got_partial(void)

/*Called by FINDER fcn main.c. -- not relvant to SCOTT*/

{
  boolean result = true;
  dynstring s;

  dynstring_initialise(&s, "");

  source = stdin;
  if (!piping) {
    result = false;
  } else {
    get_word(&s);
    if (!strcmp(s.c_str,"end")) {
      result = false;
    } else {
      if (strcmp(s.c_str,"model")) {
        fprintf(WARNING_STREAM,"\n\"%s\" found where \"model\" expected\n",s.c_str);
        result = false;
      } else {
        while (got_function());
      }
    }
  }

  return result;
}/* got_partial() */



/*------------------------------------------------------------------------*/
boolean got_function()

/* Called by got_partial(). -- not relvant to SCOTT*/

{
  dynstring s;
  int theArguments[ARGMAX];
  int i;
  functor f = NULL;
  kind k = NULL;
  int count;

  dynstring_initialise(&s, "");

  get_word(&s);
  if (s.c_str[0] == '.') {
    dynstring_free(&s);
    return false;
  }
  if (!strcmp(s.c_str,"size")) {
    get_word(&s);
    if ((i = lex_pos(s.c_str)) != -1) {
      if (lexicon[i].iskind)
	k = sort+lexicon[i].offset;
      else skipout("Incorrect sort name transfer",1);
    }
    k->siz = get_int();
    if (k->siz < k->sizmin)
      k->sizmin = k->siz;
    if (k->siz > k->sizmax)
      k->sizmax = k->siz;
    dynstring_free(&s);
    return true;
  }
  if ((i = lex_pos(s.c_str)) != -1) {
    if (!lexicon[i].iskind) 
      f = fn+lexicon[i].offset;
    else skipout("Incorrect function name transfer",1);
  }

  if (f->adicity) count = 2; else count = 1;
  for (i=0; i<f->adicity; i++) count *= (f->arg_type[i]->siz);
  free_table(f);
  f->table = (int*) malloc(sizeof(int) * count);

  if (f->adicity)
    f->fats[f->adicity-1] = 1;
  for (i=f->adicity-2; i>= 0; i--)
    f->fats[i] = f->fats[i+1] * (f->arg_type[i+1]->siz);

  if (!f->adicity) *(f->table) = get_int();
  else {
    *(f->table) = f->val_type->siz;
    if (f->adicity == 1) {
      for (i = 0; i < f->arg_type[0]->siz; i++)
	*(f->table + i + 1) = get_int();
    }
    else multiread(f,0,theArguments);
  }
  func_sanity_check(f);

  dynstring_free(&s);
  return true;

} /* got_function() */


/*---------------------------------------------------------------*/

/*
  To read back the values of a function ugly-printed by multiprint
*/

void multiread(functor f, int x, int theArguments[])
{
  for (theArguments[x] = 0;
       theArguments[x] < f->arg_type[x]->siz;
       theArguments[x]++) {
    if (x == f->adicity-1)
      *(f->table + getoffset(f,theArguments)) = get_int();
    else multiread(f,x+1,theArguments);
  }
}


/*------------------------------------------------------------------*/

void func_sanity_check(functor f)
{
  if (f) {
    if (!f->fname)
      inerror("Function symbol has no name",0,0);
    if (f->scope < 0 || f->scope > 10)
      inerror("Invalid scope definition for function symbol",f->fname,0);
    if ((f->injective || f->surjective) && f->adicity==0)
      inerror("Injective/surjective/all_different/one_one applied to 0-adic function",f->fname,0);
    if ((f->leftin || f->rightin || f->leftsur || f->rightsur) && f->adicity!=2)
      inerror("Row/column injective/surjective applied to non-binary function",f->fname,0);
  }
}



/*------------------------------------------------------------------------*/
void inerror_break_line(char *s1, char *s2, boolean drilldown)

     /*Called mostly from the parser.
      *Print input error, breaking line at the input point.
      *Exit with status set to current line number*/
{
  int x;

  fprintf(WARNING_STREAM,
	  "\nInput error on line %d:\n", current_line);
  for (x=0; x<input_point; x++) fprintf(WARNING_STREAM,"%c",job_line.c_str[x]);
  fprintf(WARNING_STREAM,"\n");
  for (x=0; x<input_point; x++) fprintf(WARNING_STREAM," ");
  fprintf(WARNING_STREAM,"%s\n\n",job_line.c_str+input_point);
  fprintf(WARNING_STREAM,
	  "%s %s\n\n",
	  s1,
	  s2);
  fflush(WARNING_STREAM);
  if (!drilldown || !setting.drilldown)
    exit(current_line);
}


/*------------------------------------------------------------------------*/
void inerror(char *s1, char *s2, boolean drilldown)

     /*Called by many functions in input.c, FINDER.c, parser.c and more.
      
      *Print error message without line break and exit as above.

      *In cases where there's a drill-down analysis of the error,
      *return rather than exiting. The further function will
      *standardly cause the exit.*/
{
  fprintf(WARNING_STREAM,
	  "\nInput error on line %d: %s\n%s %s\n\n\n", 
	  current_line, job_line.c_str,s1,s2);
  fflush(WARNING_STREAM);
  if (!drilldown || !setting.drilldown)
    exit(current_line);
}

void inwarning(char *s1, char *s2)

     /*Called by verb_spec(), verb_spec_spec(), get_set(), func_spec(), 
      *set_up_type(), sort_spec(), and get_key_word() (all in this file).
      
      *Print input warning but continue jobs.*/
{
  fprintf(WARNING_STREAM,
	  "\nWarning: possible error on line %d: %s\n%s %s\n\n\n", 
	  current_line, job_line.c_str,s1,s2);
  fflush(WARNING_STREAM);
}

/*================================ END ===================================*/
