/*
 * Contains output functions for FINDER. Some of these are called from
 * fvntr.c and other places that may not be part of FINDER, if they
 * are used in MaGIC or elsewhere, do these are separated from the
 * FINDER-intrinsic functions. Some of the functions called from
 * fvntr.c which make no sense outside FINDER are provided as stubs.
 */

#ifdef IN_FINDER
#include "FINDER.h"
#else
#include "Fvntr.h"
#endif

#ifdef SCOTT
#include "SFshared_protos.h" // protos shared by SCOTT and FINDER
#define num_theories get_num_maxisets()
#else
#define num_theories 1
#endif

/*
 * Utilities to send a string without parameters to a file or an ifile
 */

#define xprintf1(f,i,format,x) { \
  if (f) fprintf(f,format,x); \
  if (i) iprintf(i,format,x); }

#define xprintf2(f,i,format,x,y) {		\
    if (f) fprintf(f,format,x,y);		\
    if (i) iprintf(i,format,x,y); }

void xprint(FILE *f, ifile i, char *s)
// Print just the string
{
  if (f)
    fprintf(f,"%s",s);
  if (i)
    iprintf(i,"%s",s);
}

void xprintfc(FILE *f, ifile i, const char *format, char c)
// Print the char in a string context. Format must contain exactly one "%c".
{
  if (f)
    fprintf(f,format,c);
  if (i)
    iprintf(i,format,c);
}

void xprintfd(FILE *f, ifile i, const char *format, int x)
// Print the int in a string context. Format must contain exactly one "%d".
{
  if (f)
    fprintf(f,format,x);
  if (i)
    iprintf(i,format,x);
}

void xprintfld(FILE *f, ifile i, const char *format, long int x)
// As above, for long ints
{
  if (f)
    fprintf(f,format,x);
  if (i)
    iprintf(i,format,x);
}

void xprintff(FILE *f, ifile i, const char *format, double x)
// Print the number in a string context. Format must contain exactly
// one "%f or other conversion code for a float or double".
{
  if (f)
    fprintf(f,format,x);
  if (i)
    iprintf(i,format,x);
}

void xprintfs(FILE *f, ifile i, const char *format, char *s)
// Print the string in a string context. Format must contain exactly one "%s".
{
  if (f)
    fprintf(f,format,s);
  if (i)
    iprintf(i,format,s);
}

void xprintfsd(FILE *f, ifile i, const char *format, char *s, int x)
// Print the string and int in a string context. Format must contain
// exactly one "%s" and one "%d" in that order.
{
  if (f)
    fprintf(f,format,s,x);
  if (i)
    iprintf(i,format,s,x);
}

void xprintfsf(FILE *f, ifile i, const char *format, char *s, double x)
// Print the string and number in a string context. Format must
// contain exactly one "%s" and one "%f" in that order.
{
  if (f)
    fprintf(f,format,s,x);
  if (i)
    iprintf(i,format,s,x);
}


#ifndef IN_FINDER
void ref_size_report(void)                                  { /* stub */ }
void dump_refs(int rows, int columns)                       { /* stub */ }
void cgv_state_report(trs T, cell c, int val)               { /* stub */ }
void dump_state(set domains[], trs T)                       { /* stub */ }
void print_cgv_set(set domain, int maxval)                  { /* stub */ }
void print_domains(set possibles[], set impossibles[])      { /* stub */ }
void print_val_of_cell(FILE *fil, ifile ifil, int x, int v) { /* stub */ }
void print_un_vec_pos(int x)                                { /* stub */ }
void print_best_model(int info[], trs T)                    { /* stub */ }
#else


/*
*	These two are as trivial as they come.
*/

void print_the_job(void)
{
  xprint_the_job(setting.outf,setting.outi);
}

void print_job(void)
{
  xprint_job(setting.outf,setting.outi);
}




/*
*	These is the same as fprint_job and iprint_job except that the
*	job is printed in full regardless of the verbosity
*	settings. They are not called from within FINDER.
*/

void xprint_the_job(FILE *fil, ifile ifil)
{
  boolean tmp;

  tmp = setting.pm.verbosity[pjob];
  setting.pm.verbosity[pjob] = full;
  xprint_job(fil,ifil);
  setting.pm.verbosity[pjob] = tmp;
}


/*
*	Send the job description to fil, in FINDER input format.
*/

void xprint_job(FILE *fil, ifile ifil)
{
  if (!setting.pm.verbosity[pjob]) return;
  print_sorts(fil,ifil);
  print_functions(fil,ifil);
  if (setting.pm.verbosity[pjob] == full) {
    print_clauses(fil,ifil);
    print_setting(fil,ifil);
  }
  xprint(fil,ifil,"\nend\n");
}



/*
*	First obvious subroutine of fprint_job.
*/

void print_sorts(FILE *fil, ifile ifil)
{
  int i;

  for (i = 0; sort[i].sname; i++) {
    xprint(fil,ifil, "sort ");
    print_one_sort(fil,ifil,sort+i);
  }
}


/*
*	Exactly what it says.
*/

void print_one_sort(FILE *fil, ifile ifil, kind k)
{
  name_list ep;

  if (k == Bool) {
    xprint(fil,ifil, " bool {}\n");
    return;
  }
  if (k->piped) {
    if (fil) fprintf(fil," %s { %c piped  ", k->sname, '%');
    if (ifil) iprintf(ifil," %s { %c piped  ", k->sname, '%');
  }
  else
    xprintfs(fil,ifil, " %s { ", k->sname);
  if (k->canonical && (k!=Integer)) {
    xprint(fil,ifil, "enum: ");
    for (ep=k->canonical; ep; ep=ep->nextname) {
      if (ep != k->canonical) xprint(fil,ifil, ", ");
      xprint(fil,ifil, ep->name);
    }
    xprint(fil,ifil, ".");
  }
  else if (k->sizmin == k->sizmax)
    xprintfd(fil,ifil,"cardinality = %d", k->sizmin);
  else {
    if (fil) fprintf(fil,"cardinality > %d  cardinality < %d",
		     k->sizmin-1, k->sizmax+1);
    if (ifil) iprintf(ifil,"cardinality > %d  cardinality < %d",
		     k->sizmin-1, k->sizmax+1);
  }
  xprint(fil,ifil, " }\n");
}



/*
*	Second obvious subroutine of fprint_job.
*/

void print_functions(FILE *fil, ifile ifil)
{
  functor f;

  for (f = FN; f; f = f->nextfunct) {
    if (setting.pm.pretty || f->pmode) {
      switch (setting.function_spec_mode) {
      case math:
	xprint(fil,ifil, "function ");
	break;
      case comp:
	xprint(fil,ifil, "vocabulary ");
	break;
      }
      print_one_function(fil,ifil, f);
    }
  }
}


/*
*	Describing a function symbol is more complex than describing a sort,
*	but conceptually it is no deeper.
*
*       TODO: if in comp spec mode, print accordingly
*/

void print_one_function(FILE *fil, ifile ifil, functor f)
{
  spec_type mode;
  int i;

  mode = setting.function_spec_mode;
  switch(mode) {
  case math:
    xprintfs(fil,ifil, "%s: ", f->fname);
    for (i=0; i<f->adicity; i++) {
      if (i) xprint(fil,ifil,", ");
      if (f->arg_type[i])
	xprint(fil,ifil, f->arg_type[i]->sname);
      else xprint(fil,ifil, "??");
    }
    if (f->adicity) xprint(fil,ifil, " -> ");
    xprint(fil,ifil, f->val_type->sname);
    if (!f->adicity) xprint(fil,ifil, ".");
    break;
  case comp:
    if (f->adicity == 0)
      xprint(fil,ifil, "name ");
    else if (f->val_type == Bool)
      xprint(fil,ifil, "predicate ");
    else
      xprint(fil,ifil, "function ");
    xprintfs(fil,ifil, "%s ", f->fname);
    if (f->adicity) {
      xprint(fil,ifil, "(");
      for (i=0; i<f->adicity; i++) {
	if (i) xprint(fil,ifil, ",");
	if (f->arg_type[i])
	  xprint(fil,ifil, f->arg_type[i]->sname);
	else xprint(fil,ifil, "??");
      }
      xprint(fil,ifil, ")");
    }
    if (f->val_type != Bool || f->adicity == 0) {
      xprintfs(fil,ifil, " : %s",f->val_type->sname);
    }
    break;
  }
  xprint(fil,ifil, " {\n");
  if (f->total) xprint(fil,ifil, "  total\n");
  else xprint(fil,ifil, "  partial\n");
  if (f->injective && f->surjective) xprint(fil,ifil, "  bijective\n");
  else {
    if (f->injective) xprint(fil,ifil, "  injective\n");
    if (f->surjective) xprint(fil,ifil, "  surjective\n");
  }	
  if (f->leftin && f->leftsur) xprint(fil,ifil, "  column_bijective\n");
  else {
    if (f->leftin) xprint(fil,ifil, "  column_injective\n");
    if (f->leftsur) xprint(fil,ifil, "  column_surjective\n");
  }
  if (f->rightin && f->rightsur) xprint(fil,ifil, "  row_bijective\n");
  else {
    if (f->rightin) xprint(fil,ifil, "  row_injective\n");
    if (f->rightsur) xprint(fil,ifil, "  row_surjective\n");
  }
  if (f->comm) xprint(fil,ifil, "  commutative\n");
  if (f->cut_it) xprint(fil,ifil, "  cut\n");
  if (f->nop) xprint(fil,ifil, "  no_priority\n");
  if (!f->labels) xprint(fil,ifil, "  no_labels\n");
  xprintfd(fil,ifil, "  scope: %d\n", f->scope);
  if (f->pmode == null) xprint(fil,ifil, "  hidden\n");
  else if (f->pmode == brief) 
    xprint(fil,ifil, "  print: brief\nt");
  else	xprint(fil,ifil, "  print: full\n");
  xprint(fil,ifil, "}\n");
}


/*
*	Third obvious subroutine of fprint_job.
*/

void print_clauses(FILE *fil, ifile ifil)
{
  clause c;

  for (c=TH[current_theory]; c; c=c->nextclause) {
    print_one_clause(fil,ifil,c,false);
  }
}


/*
 * Suppress prefixing of quantifiers, etc
 */

void cprint_one_clause(FILE *fil, ifile ifil, clause c, boolean plain)
{
  boolean b;

  b = setting.single_literal;
  setting.single_literal = false;
  print_one_clause(fil,ifil,c,plain);
  setting.single_literal = b;
}




/*
*	Printing a clause is simple. Add the weight, softweight and position.
*/

void print_one_clause(FILE *fil, ifile ifil, clause c, boolean plain)
{
  wff_stack wf;
  
  if (!fil && !ifil) fil = stdout;
  if (!plain)
    xprintfd(fil,ifil, "clause (%d) { ",c->softweight);
  if (setting.single_literal)
    FOFprint_clause(fil,ifil,c);
  else {
    for (wf=c->prem; wf; wf=wf->nxt) {
      if (wf != c->prem) xprint(fil,ifil, ", ");
      outwff(fil,ifil,wf->w);
    }
    if (c->prem) xprint(fil,ifil, "  ->  ");
    for (wf=c->conc; wf; wf=wf->nxt) {
      if (wf != c->conc) xprint(fil,ifil, "; ");
      outwff(fil,ifil,wf->w);
    }
    if (!c->conc) xprint(fil,ifil, "FALSE");
  }
  if (plain)
    xprint(fil,ifil, ".\n");
  else {
    xprint(fil,ifil, ". }");
    if (fil) fprintf(fil,"  %c %d %d %d %d %d %d ",
		     '%', c->varcount, c->weight, c->tuples,
		     c->softweight, c->cid, c->pid);
    if (ifil) iprintf(ifil,"  %c %d %d %d %d %d %d ",
		      '%', c->varcount, c->weight, c->tuples,
		      c->softweight, c->cid, c->pid);
    /*
     * Print variables and their sorts for reference
     */
    for (wf=c->varlist; wf; wf=wf->nxt)
      xprintfs(fil,ifil, " %s ",wf->w->sym->fname);

    if (c->tautology)
      xprint(fil,ifil, " TAUTOLOGY");
    xprint(fil,ifil, "\n");
  }
}



void FOFprint_clause (FILE *fil, ifile ifil, clause c)
// Subroutine to print the formula if FOF form is set
{
  int nprems, nconcs;
  wff_stack wf;

  for (nprems=0,wf=c->prem; wf; wf=wf->nxt) nprems++; 
  for (nconcs=0,wf=c->conc; wf; wf=wf->nxt) nconcs++;
  if (nprems+nconcs == 0) return;
  if (c->varlist) {
    for (wf=c->varlist; wf; wf=wf->nxt) {
      xprint(fil,ifil, "ALL ");
      print_without_type(fil,ifil,wf->w);
      xprint(fil,ifil, " ");
    }
    xprint(fil,ifil, "(");
  }
  if (nprems == 0)
    for (wf=c->conc; wf; wf=wf->nxt) {
      outwff(fil,ifil,wf->w);
      if (wf->nxt)
	xprint(fil,ifil, " OR ");
    }
  else if (nconcs == 0)
    for (wf=c->prem; wf; wf=wf->nxt) {
      xprint(fil,ifil, "NOT(");
      outwff(fil,ifil,wf->w);
      xprint(fil,ifil, ")");
      if (wf->nxt)
	xprint(fil,ifil, " OR ");
    }
  else {
    if (nprems > 1)
      xprint(fil,ifil, "(");
    for (wf=c->prem; wf; wf=wf->nxt) {
      outwff(fil,ifil,wf->w);
      if (wf->nxt)
	xprint(fil,ifil, " AND ");
    }
    if (nprems > 1)
      xprint(fil,ifil, ")");
    xprint(fil,ifil, " IMP ");

    if (nconcs > 1)
      xprint(fil,ifil, "(");
    for (wf=c->conc; wf; wf=wf->nxt) {
      outwff(fil,ifil,wf->w);
      if (wf->nxt)
	xprint(fil,ifil, " OR ");
    }
    if (nconcs > 1)
      xprint(fil,ifil, ")");
  }
  if (c->varlist)
    xprint(fil,ifil, ")");
}



void printwff(wff w) {outwff(WARNING_STREAM,0,w);}

/*
*	Recursively print a formula, in Polish except for =, <, >, +. -.
*/

void outwff(FILE *fil, ifile ifil, wff w)
{
  if (!w) return;
  if (!fil && !ifil) fil = stdout;
  if (strchr("=><+-",w->sym->fname[0])) {
    outsubwff(fil,ifil,w->sb[0]);
    // TODO: strip subscripts off variables
    xprintfc(fil,ifil," %c ",w->sym->fname[0]);
    outsubwff(fil,ifil,w->sb[1]);
  }
  else if (!strcmp("!=",w->sym->fname)) {
    outsubwff(fil,ifil,w->sb[0]);
    // TODO: strip subscripts off variables
    xprint(fil,ifil," != ");
    outsubwff(fil,ifil,w->sb[1]);
  }
  else
    outsubwff(fil,ifil,w);
}


void print_without_type(FILE *fil, ifile ifil, wff w)
{
  char s[LEXLENGTH];

  strcpy(s,w->sym->fname);
  if (strchr(s,':'))
    *(strchr(s,':')) = 0;
  xprint(fil,ifil, s);
}


void outsubwff(FILE *fil, ifile ifil, wff w)
{
  int i;

  if (!w) return;
  if (strchr("=><+-",w->sym->fname[0]) || !strcmp("!=",w->sym->fname)) {
    xprint(fil,ifil, "(");
    outsubwff(fil,ifil,w->sb[0]);
    if (!strcmp("!=",w->sym->fname))
      xprint(fil,ifil," != ");
    else
      xprintfc(fil,ifil," %c ",w->sym->fname[0]);
    outsubwff(fil,ifil,w->sb[1]);
    xprint(fil,ifil, ")");
  }
  else if (w->sym->category == Quant) {
    xprintfs(fil,ifil, "%s ", w->sym->fname);
    outsubwff(fil,ifil,w->sb[0]);
    xprint(fil,ifil, " ");
    outsubwff(fil,ifil,w->sb[1]);
  }
  else {
    if (setting.single_literal)
      print_without_type(fil,ifil,w);
    else
      xprintfs(fil,ifil, "%s", w->sym->fname);
    if (w->sym->adicity) {
      xprint(fil,ifil, "(");
      for (i=0; i<w->sym->adicity; i++) {
	if (i)
	  xprint(fil,ifil, ",");
	outsubwff(fil,ifil,w->sb[i]);
      }
      xprint(fil,ifil, ")");
    }
  }
  if (fil) fflush(fil);
}


void indentwff(FILE *f, wff w, int i, boolean suppress)
// Print a wff for disgnostic purposes, with binary operators in infix,
// using indentation (i spaces) to help make the structure clear
// At present, always goes to warning stream, but that may need to change
{
  int x;

  if (!w) return;

  if (!suppress)
    print_indent(f,i);
  if (w->sym->adicity == 0)
    fprintf(f,"%s\n",w->sym->fname);
  else if (w->sym->category == Quant) {
    fprintf(f,"%s %s (\n",w->sym->fname,w->sb[0]->sym->fname);
    indentwff(f,w->sb[1],i+2,false);
    print_indent(f,i);
    fprintf(f,")\n");
  }
  else if (w->sym->adicity == 2) {
    if (w->sb[0]->sym->adicity == 2 && w->sb[0]->sym->category != Quant) {
      fprintf(f,"( ");
      indentwff(f,w->sb[0],i+2,true);
      print_indent(f,i);
      fprintf(f,")\n");
    }
    else
      indentwff(f,w->sb[0],i,true);
    print_indent(f,i);
    fprintf(f,"%s\n",w->sym->fname);
    if (w->sb[1]->sym->adicity == 2&& w->sb[0]->sym->category != Quant) {
      print_indent(f,i);
      fprintf(f,"( ");
      indentwff(f,w->sb[1],i+2,true);
      print_indent(f,i);
      fprintf(f,")\n");
    }
    else
      indentwff(f,w->sb[1],i,false);
  }
  else {
    fprintf(f,"%s (\n",w->sym->fname);
    for (x=0; x<w->sym->adicity; x++) {
      indentwff(f,w->sb[x],i+2,false);
      print_indent(f,i);
      if (x < w->sym->adicity-1)
	fprintf(f,"  ,\n");
      else
	fprintf(f,")\n");
    }
  }
}

void print_indent(FILE *f, int i)
{
  int x;
  
  for (x=0; x<i; x++)
    fprintf(f," ");
  fflush(f);
}



/*
*	Fourth obvious subroutine of fprint_job.
*/

void print_setting(FILE *fil, ifile ifil)
{
  xprint(fil,ifil, "\nsetting {\n");
  prints_int(fil,ifil, "solutions", setting.max_models);
  prints_int(fil,ifil, "backtracks", setting.max_backtracks);
  prints_int(fil,ifil, "ref_length", setting.maxref);
  prints_int(fil,ifil, "test_ref_length", setting.test_maxref);
  prints_int(fil,ifil, "time_limit", setting.maxtime);
  prints_int(fil,ifil, "pre_test", setting.pretest_limit);
  prints_int(fil,ifil, "random", (setting.random? 1: 0));
  prints_bool(fil,ifil, "primary", "secondary", setting.Nosecs);
  prints_bool(fil,ifil, "reverse_test", "obverse_test", setting.invert_change);
  prints_bool(fil,ifil, "pretty", "ugly", setting.pm.pretty);
  prints_bool(fil,ifil, "noisy", "quiet", setting.noisy);
  prints_bool(fil,ifil, "time_stamp", "no_time_stamp", setting.timestamps);
  prints_bool(fil,ifil, "dump_depth", "no_dump_depth", setting.dumpdepth);
  prints_bool(fil,ifil, "awnty", "no_awnty", setting.awnty_enabled);
  prints_bool(fil,ifil, "debug", "no_debug", setting.debug_enabled);
  prints_bool(fil,ifil, "tops_off", "tops_on", setting.topsoff);
  xprintfc(fil,ifil, "%c ",'%');           // Next setting is done via the command line
  prints_bool(fil,ifil, "FOF_form", "clause_form", setting.single_literal);
  prints_bool(fil,ifil, "report_breaks", "silent_breaks", setting.report_breaks);
  prints_bool(fil,ifil, "test_by_vars", "test_by_weight", setting.vars_over_wt);
  prints_bool(fil,ifil, "no_process", "process", setting.parse_only);
  prints_bool(fil,ifil, "no_search", "search", setting.clausify_only);
  prints_bool(fil,ifil, "local_search", "global_search", setting.search_is_local);
  xprint(fil,ifil, "\tstack: ");
  if (setting.batches == S_SMALL) xprint(fil,ifil, "small\n");
  else if (setting.batches == S_MEDIUM) xprint(fil,ifil, "medium\n");
  else if (setting.batches == S_LARGE) xprint(fil,ifil, "large\n");
  else if (setting.batches == S_MAXIMAL) xprint(fil,ifil, "maximal\n");
  else xprintfd(fil,ifil, "%d\n", setting.batches);
  xprint(fil,ifil, "\tverbosity {\n\t");
  prints_verb(fil,ifil, "models", pmodels);
  prints_verb(fil,ifil, "job", pjob);
  prints_verb(fil,ifil, "stats", pstats);
  xprint(fil,ifil, "}\n}\n");
}




/*
*	There now follow the subroutines called by print_set.
*/

void prints_int(FILE *fil, ifile ifil, char *s, int x)
{
  if (x)
    xprintfsd(fil,ifil,"\t%s %d\n", s, x);
}


void prints_bool(FILE *fil, ifile ifil, char *strue, char *sfalse, boolean b)
{
  xprintfs(fil,ifil, "\t%s\n", b? strue: sfalse);
}


void prints_verb(FILE *fil, ifile ifil, char *s, print_mode pmd)
{
  prints p;

  p = setting.pm.verbosity[pmd];
  switch(p) {
  case null:
    xprintfs(fil,ifil, "\t%s: none\n\t", s);
    return;
  case brief:
    xprintfs(fil,ifil, "\t%s: brief\n\t", s);
    return;
  case full:
    xprintfs(fil,ifil, "\t%s: full\n\t", s);
    return;
  }
  skipout("Strange verbosity setting",1);
}


/*
=============================================================================
*/


/*
*	Print_model is called when there is a new model to report. It checks
*	for error conditions such as reporting the same model twice and 
*	as a side effect maintains the vector last_good. It also updates the
*	global counter good.
*
*	There are two formats, pretty and ugly (human-readable and machine-
*	readable). In addition, there are two print modes. Brief means print
*	only when there is a change to report, and Full means print no matter 
*	what. There is also None, which is self-explanatory. There is a 
*	general verbosity setting and a local one for each function. The 
*	lower of these two over-rides the higher.
*
*	Before the real thing starts, there are some trivia as in the 
*	above job-printing case.
*/

void print_the_model(void)
{
  xprint_the_model(setting.outf,setting.outi);
}

void xprint_the_model(FILE *fil, ifile ifil)
{
  prints tmp;

  tmp = setting.pm.verbosity[pmodels];
  setting.pm.verbosity[pmodels] = full;
  xprint_model(fil,ifil);
  setting.pm.verbosity[pmodels] = tmp;
}

void print_model(void)
{
  xprint_model(setting.outf,setting.outi);
}

void xprint_model(FILE *fil, ifile ifil)
{
  kind k;
  functor f;
  int timenow;
  boolean Full, Brief, header;
  char headline[100];              // We know the maximum length!
  int theArguments[ARGMAX];
  int i;

  header = setting.print_headers;
  if (setting.pm.verbosity[pmodels] && !setting.pm.pretty)
    xprint(fil,ifil,"model\n");
  
  if (firstmodel) {
    firstmodel = false;
    if (setting.pm.verbosity[pmodels] && !setting.pm.pretty) {
      for (k = sort+1; k->sname; k++)
	xprintfsd(fil,ifil,"size %s %d\n", k->sname, k->siz);
    }
  }

  if (setting.noisy) fprintf(WARNING_STREAM,"\007");      // Bell
  if (!(setting.pm.verbosity[pmodels])) return;
  for (f = FN; f; f = f->nextfunct) 
    if (!negative || f->piped) {
  
  Full = (f->pmode==full && setting.pm.verbosity[pmodels]==full);

  Brief = ((f->pmode==full && setting.pm.verbosity[pmodels]==brief)
	       || (f->pmode==brief && setting.pm.verbosity[pmodels]));

      if (Full || (Brief && (f->piped || f->fcc <= first_change))) {

	if (header && setting.pm.pretty) {
	  if (num_theories > 1) {
	    if (setting.timestamps) {
	      timenow = timestamp()-begin_time;
	      if (theBadness) {
		if (fil)
		  fprintf(fil,
			  "\nModel %d.%d [ %d %d %d ] (%3.2f seconds, %d Inserts)\n",
			  current_theory,
			  updates[current_theory],
			  theBadness,
			  theTime-begin_time,
			  theWork,
			  (timenow*1.0)/theClockTick,
			  workNow);
		if (ifil)
		  iprintf(ifil,
			  "\nModel %d.%d [ %d %d %d ] (%3.2f seconds, %d Inserts)\n",
			  current_theory,
			  updates[current_theory],
			  theBadness,
			  theTime-begin_time,
			  theWork,
			  (timenow*1.0)/theClockTick,
			  workNow);
	      }
	      else {
		if (fil)
		  fprintf(fil,
			  "\nModel %d.%d (%3.2f seconds, %d Inserts)\n",
			  current_theory,
			  updates[current_theory],
			  (timenow*1.0)/theClockTick,
			  workNow);
		if (ifil)
		  iprintf(ifil,
			  "\nModel %d.%d (%3.2f seconds, %d Inserts)\n",
			  current_theory,
			  updates[current_theory],
			  (timenow*1.0)/theClockTick,
			  workNow);
	      }
	    }
	    else {
	      if (theBadness) {
		if (fil)
		  fprintf(fil,"\nModel %d.%d [ %d %d %d ] (%d Inserts)\n",
			  current_theory,
			  updates[current_theory],
			  theBadness,
			  theTime-begin_time,
			  theWork,
			  workNow);
		if (ifil)
		  iprintf(ifil,"\nModel %d.%d [ %d %d %d ] (%d Inserts)\n",
			  current_theory,
			  updates[current_theory],
			  theBadness,
			  theTime-begin_time,
			  theWork,
			  workNow);
	      }
	      else {
		if (fil)
		  fprintf(fil,"\nModel %d.%d\n",
			  current_theory,
			  updates[current_theory]);
		if (ifil)
		  iprintf(ifil,"\nModel %d.%d\n",
			  current_theory,
			  updates[current_theory]);
	      }
	    }
	  }
	  else {
	    if (setting.timestamps) {
	      timenow = timestamp()-begin_time;
	      if (theBadness)
		sprintf(headline,
			"Model %d [ %d %d %d ] (%3.2f seconds, %d Inserts)",
			updates[0],
			theBadness,
			theTime-begin_time,
			theWork,
			(timenow*1.0)/theClockTick,
			workNow);
	      else
		sprintf(headline,
			"Model %d (%3.2f seconds, %d Inserts)",
			updates[0],
			(timenow*1.0)/theClockTick,
			workNow);
	    }
	    else {
	      if (theBadness) {
#ifdef WORLDWIDE
		sprintf(headline,"Approximate model");
#else
		sprintf(headline,"Model %d [ %d %d %d ] (%d Inserts)",
			updates[0],
			theBadness,
			theTime-begin_time,
			theWork,
			workNow);
#endif
	      }
	      else
		sprintf(headline,"Model %d",updates[0]);
	    }
	  }
	  xprintfs(fil,ifil,"\n%s\n",headline);
	  header = false;
	}

	for (i=0; i<ARGMAX; i++) theArguments[i] = 0;

	if (f->adicity > 1)
	  multiprint(fil,ifil,f,0,theArguments);
	else switch(f->adicity) {
	case 0:
	  zeroprint(fil,ifil, f);
	  break;
	case 1:
	  oneprint(fil,ifil, f);
	  break;
	}
	xprint(fil,ifil,"\n");
      }
    }
  if (!setting.pm.pretty) xprint(fil,ifil,".\n");
  else if (!header) xprint(fil,ifil,"\n");

  if (theBadness && setting.print_unsat)
    print_violated_clauses(fil,ifil);
}



/*
*	Print a zero-adic function.
*/

void zeroprint(FILE *fil, ifile ifil, functor f)
{
    if (setting.pm.pretty) {
      xprintfs(fil,ifil,"\n%s = ", f->fname);
      printvalue(fil,ifil, f->val_type, *(f->table), 0, 0);
    }
    else xprintfsf(fil,ifil, "%s %d", f->fname, *(f->table));
}


/*
*	Print a monadic function.
*/

void oneprint(FILE *fil, ifile ifil, functor f)
{
  int indentation;
  int *columnwidth = 0;
  int i, j;

  if (setting.pm.pretty) {
    indentation = get_monadic_indentation(f);
    columnwidth = get_monadic_columnwidth(f);

    if (f->labels) {
      xprint(fil,ifil,"\n");
      indent(fil,ifil,indentation-strlen(f->arg_type[0]->sname));
      xprintfs(fil,ifil, "%s | ", f->arg_type[0]->sname);
      for (i = 0; i < f->arg_type[0]->siz; i++)
	printvalue(fil,ifil, f->arg_type[0], i, columnwidth[i], 0);
      xprint(fil,ifil,"\n");
      for (i = 0; i <= indentation; i++) xprint(fil,ifil,"-");
      xprint(fil,ifil,"+-");
      for (j = 0; j < f->arg_type[0]->siz; j++) {
	for (i = 0; i < columnwidth[j]-1; i++)
	  xprint(fil,ifil,"-");
	if (j < f->arg_type[0]->siz-1)
	  xprint(fil,ifil,"--");
      }
      xprint(fil,ifil,"\n");
      indent(fil,ifil,indentation-strlen(f->fname));
      xprintfs(fil,ifil, "%s | ", f->fname);
    }
    for (i = 0; i < f->arg_type[0]->siz; i++)
      printvalue(fil,ifil, f->val_type, f->table[i+1], columnwidth[i], 0);
  }
  else {                             // Ugly print
    xprintfs(fil,ifil, "%s ", f->fname);
    for (i = 0; i < f->arg_type[0]->siz; i++)
      xprintfd(fil,ifil, " %d", *(f->table+i+1));
  }
  if (columnwidth) free(columnwidth);
}



/*
 * Print a function of adicity greater than 1
 */

void multiprint(FILE *fil, ifile ifil, functor f, int x, int theArguments[])
{
  int v;
  dynstring stuff;
  int indentation = 0;
  int *columnwidth = 0;
  int i,j;

  dynstring_initialise(&stuff, "");

  if (!setting.pm.pretty) {
    if (!x) xprintfs(fil,ifil, "%s ", f->fname);
    for (theArguments[x] = 0;
	 theArguments[x] < f->arg_type[x]->siz;
	 theArguments[x]++) {
      if (x == f->adicity-1)
	xprintfd(fil,ifil, " %d", *(f->table + getoffset(f,theArguments)));
      else multiprint(fil,ifil,f,x+1,theArguments);
    }
  }
  else {
    if (x < f->adicity-2) {
      for (theArguments[x]=0;
	   theArguments[x] < f->arg_type[x]->siz;
	   theArguments[x]++)
	multiprint(fil,ifil,f,x+1,theArguments);
    }
    else {
      columnwidth = get_column_width(f,theArguments);
      if (f->labels) {
	indentation = get_indent(f,theArguments);
	xprint(fil,ifil, "\n");
	if (f->adicity == 2) {
	  indent(fil,ifil, indentation - (strlen(f->fname)+1));
	  xprintfs(fil,ifil, "%s ", f->fname);
	}
	else {
	  i = strlen(f->fname)+5;
	  for (v=0; v<f->adicity-2; v++) {
	    if (v) i += 1;
	    i += f->arg_type[v]->len[theArguments[v]];
	  }
	  indent(fil,ifil,indentation-i);
	  xprintfs(fil,ifil, "%s(", f->fname);
	  for (v=0; v<f->adicity-2; v++) {
	    if (v) xprint(fil,ifil,",");
	    get_the_name(f->arg_type[v],theArguments[v],&stuff);
	    xprintfs(fil,ifil, "%s", stuff.c_str);
	  }
	  xprint(fil,ifil, ",_) ");
	}
	xprint(fil,ifil,"|");
	for (i=0; i < f->arg_type[x+1]->siz; i++) {
	  get_the_name(f->arg_type[x+1],i,&stuff);
	  xprintfs(fil,ifil, " %s",stuff.c_str);
	  indent(fil,ifil, columnwidth[i]- stuff.length);
	}
	xprint(fil,ifil,"\n");
	for (i=0; i<indentation; i++) xprint(fil,ifil,"-");
	xprint(fil,ifil,"+");
	for (i=0; i < f->arg_type[x+1]->siz; i++)
	  for (j=0; j<=columnwidth[i]; j++)
	    xprint(fil,ifil,"-");
	xprint(fil,ifil,"\n");
      }
      for (theArguments[x] = 0;
	   theArguments[x] < f->arg_type[x]->siz;
	   theArguments[x]++) {
	if (f->labels) {
	  get_the_name(f->arg_type[x],theArguments[x],&stuff);
	  indent(fil,ifil, indentation-(stuff.length+1));
	  xprintfs(fil,ifil,"%s |",stuff.c_str);
	}
	else indent(fil,ifil,indentation+1);
	for (theArguments[x+1] = 0;
	     theArguments[x+1] < f->arg_type[x+1]->siz;
	     theArguments[x+1]++) {
	  get_the_value(f,theArguments,&stuff);
	  xprintfs(fil,ifil, " %s", stuff.c_str);
	  if (theArguments[x+1] < f->arg_type[x+1]->siz)
	    indent(fil,ifil, columnwidth[theArguments[x+1]]-stuff.length);
	}
	xprint(fil,ifil,"\n");
      }
    }
  }

  if (columnwidth) free(columnwidth);
  dynstring_free(&stuff);
}


void indent(FILE *fil, ifile ifil, int x)
{
  int i;

  for (i=0; i<x; i++) xprint(fil,ifil," ");
}



int get_monadic_indentation(functor f)
{
  int len1, len2;

  len1 = strlen(f->arg_type[0]->sname);
  len2 = strlen(f->fname);
  if (len1 > len2) return len1;
  return len2;
}


int *get_monadic_columnwidth(functor f)
{
  int x;
  int siz = f->arg_type[0]->siz;
  int *cw;

  cw = (int*) malloc(siz * sizeof(int));
  for (x=0; x<siz; x++) {
    cw[x] = 2;
    if (f->arg_type[0]->len[x] > cw[x])
      cw[x] = f->arg_type[0]->len[x];
    if (f->table[x+1] < f->val_type->siz &&
	f->val_type->len[f->table[x+1]] > cw[x])
      cw[x] = f->val_type->len[f->table[x+1]];
    cw[x]++;
  }
  return cw;
}


int get_indent(functor f, int theArguments[])
{
  int oldlen;
  int newlen;
  int v;

  oldlen = 2;

  newlen = strlen(f->fname)+1;
  if (f->adicity > 2) newlen += 4;
  for (v=0; v<f->adicity-2; v++) {
    if (v) newlen += 1;
    newlen += f->arg_type[v]->len[theArguments[v]];
  }
  if (newlen > oldlen) oldlen = newlen;

  for (v=0; v < f->arg_type[f->adicity-2]->siz; v++)
    if (f->arg_type[f->adicity-2]->len[v] >= oldlen)
      oldlen = f->arg_type[f->adicity-2]->len[v]+1;
  return oldlen;
}



int *get_column_width(functor f, int theArguments[])
{
  int i,j;
  dynstring stuff;
  int columns;
  int * cw = 0;

  dynstring_initialise(&stuff, "");

  if (f->adicity > 0) {
    i = f->adicity-1;
    columns = f->arg_type[i]->siz;
    cw = (int*) malloc(columns * sizeof(int));
    for (j=0; j < columns; j++) cw[j] = 2;
    for (j=0; j < columns; j++)
      if (f->arg_type[i]->len[j] > cw[j])
	cw[j] = f->arg_type[i]->len[j];
    if (f->adicity > 1) {
      for (theArguments[i-1] = 0;
	   theArguments[i-1] < f->arg_type[i-1]->siz;
	   theArguments[i-1]++) {
	for (theArguments[i] = 0;
	     theArguments[i] < columns;
	     theArguments[i]++) {
	  get_the_value(f,theArguments,&stuff);
	  j = stuff.length;
	  if (j > cw[theArguments[i]])
	    cw[theArguments[i]] = j;
	}
      }
    }
    else {
      for (theArguments[0] = 0;
	   theArguments[0] < f->arg_type[0]->siz;
	   theArguments[0]++) {
	get_the_value(f,theArguments,&stuff);
	  j = stuff.length;
	  if (j > cw[theArguments[i]])
	    cw[theArguments[i]] = j;
      }
    }
  }

  dynstring_free(&stuff);
  return cw;
}



void get_the_value(functor f, int theArguments[], dynstring *dest)
{
  get_the_name(f->val_type,
	       *(f->table + getoffset(f,theArguments)),
	       dest);
}


void get_the_name(kind k, int x, dynstring *dest)
{
  name_list lp;
  int i;

  if (x >= k->siz) dynstring_strcpyc(dest," ");
  else if (k->canonical) {
    lp = k->canonical;
    for (i=0; i<x; i++) lp = lp->nextname;
    dynstring_strcpyc(dest,lp->name);
  }
  else
    dynstring_sprintf(dest,"%d",x);
}



/*
*	Return the length of the longest canonical name of sort k.
*/

int canonlength(kind k)
{
  int i = 0, j;

  for (j = 0; j < k->siz; j++)
    if (k->len[j] > i) i = k->len[j];
  return i;
}




/*
*	Print the canonical name of object x of kind k, padded on the
*	right with spaces up to the length len. The nullval argument
*	is the single character (default space) used to represent the
*	non-value.
*/

void printvalue(FILE *fil, ifile ifil, kind k, int x, int len, char nullval)
{
  name_list lp;
  int i, j;

  if (!nullval) nullval = ' ';
  if (x == k->siz) {
    xprintfc(fil,ifil, "%c", nullval);
    j = 1;
  }
  else if (k->canonical && setting.pm.pretty) {
    lp = k->canonical;
    for (i = 0; i < x; i++) lp = lp->nextname;
    xprintfs(fil,ifil, "%s", lp->name);
    j = strlen(lp->name);
  }
  else {
    j = 2;
    xprintfd(fil,ifil, "%-2d", x);
  }
  for (i = j; i <= len; i++) xprint(fil,ifil," ");
  if (fil) fflush(fil);
}


void print_violated_clauses(FILE *fil, ifile ifil)
{
  clause c;
  int count = 0;

  for (c=TH[current_theory]; c; c=c->nextclause)
    if (!is_good(c)) {
      count++;
      if (count == 1)
	xprint(fil,ifil, "\nViolated constraints:\n");
      print_one_clause(fil,ifil,c,true);
    }
  if (count)
    xprint(fil,ifil,"\n");
  else
    xprint(fil,ifil, "\nAll constraints are satisfied\n\n");
}



/*
============================================================================
*/


/*
*	When the FINDER job has been completed, some statistics on the run
*	are printed out. These are determined by the search controller used.
*/

void printstats(void)
{
  if (setting.pm.number_only)
    dump_report(&tr_parameters);
  else {
    xprint(setting.outf,setting.outi,"\n\n\n");
    dump_report(&tr_parameters);
    xprint(setting.outf,setting.outi,"\n");
    dump_report(&ttr_par);
    xprintff(setting.outf,setting.outi,"FINDER time: %3.2f seconds\n\n",
	     ((end_time-begin_time)*1.0) / theClockTick);
  }
}



void ref_size_report(void)
// At present, this does nothing. If it is ever re-activated, it may
// need updating to take account of possible ifiles.
{
  int i, j, k, m;
  int ctotal[RRB], rtotal[RRA], totaltotal;

  if (0) {
    totaltotal = 0;
    for (i=0; i<RRB; i++) {
      ctotal[i] = 0;
      for (j=0; j<RRA; j++)
	ctotal[i] += tr_parameters.ref_record[j][i];
    }
    for (i=0; i<RRA; i++) {
      rtotal[i] = 0;
      for (j=0; j<RRB; j++)
	rtotal[i] += tr_parameters.ref_record[i][j];
      totaltotal += rtotal[i];
    }

    for (k=RRA-1; k >= 0; k--) {
      j = 0;
      for (i=0; i<RRB; i++)
	if (tr_parameters.ref_record[k][i]) j = 1;
      if (j) break;
    }
    if (k < 0) return;
    for (m=RRB-1; m >= 0; m--) {
      j = 0;
      for (i=0; i <= k; i++)
	if (tr_parameters.ref_record[i][m]) j = 1;
      if (j) break;
    }
    if (!m) return;
    dump_refs(k+1,m+1);
  }
}




/*
*	The distribution matrix of the refutations is dumped to file
*	FINDER.refs. First the numbers of rows and columns are printed
*	along with the maximum number in the grid.
*/

void dump_refs(int rows, int columns)
// Doesn't happen at present (see ref_size_report)
{
  FILE *ref_file;
  int i, j, ccmax = 0;

  for (i = rows-1; i >= 0; i--)
    for (j = 0; j < columns; j++)
      if (tr_parameters.ref_record[i][j] > ccmax)
	ccmax = tr_parameters.ref_record[i][j];
  if (!(ref_file = fopen("FINDER.refs","w"))) return;
  fprintf(ref_file,"%d %d %d\n", rows, columns, ccmax);

  for (i = rows-1; i >= 0; i--) {
    for (j = 0; j < columns; j++)
      fprintf(ref_file, " %d", tr_parameters.ref_record[i][j]);
    fprintf(ref_file, "\n");
  }
  fclose(ref_file);
}


/*
 * For now, dump the domains. Later maybe dump a CGV batch of changes
 * since the last report
 */

void cgv_state_report(trs T, cell c, int val)
{
  set domains = 0;
  int x,y;
  functor f;
  int args[ARGMAX];

  if (T->job->vlength)
    domains = (set) malloc(T->job->vlength * sizeof(SET));
  for (x=0; x<T->job->vlength; x++) {
    S_new(T->vector[x].csiz,domains+x);
    if (T->vector[x].has_value)
      S_into(T->icomvec[x],domains+x);
    else
      for (y=0; y<T->vector[x].csiz; y++)
	if (!(T->vector[x].cvals[y].blocks))
	  S_into(y,domains+x);
  }
  printf("cgv start\n");
  dump_state(domains,T);
  if (c) {
    printf("cgv assigned ");
    f = un_vec_pos(c->vpos,args);
    printf("%s",f->fname);
    for (x=0; x<f->adicity; x++) printf(" %d",args[x]);
    printf(" %d\n",val);
  }
  else printf("cgv solution\n");
  printf("cgv end\n");
  if (domains)  {
    for (x=0; x<T->job->vlength; x++)
      S_free_contents(domains+x);
    free(domains);
  }
}


void dump_state(set domains, trs T)
{
  functor f;
  int argument[ARGMAX];
  int x,y;
  boolean real_value;

  for (f=FN; f; f=f->nextfunct)
    if (f->cgv) {
      printf("cgv %s %d ",f->fname,f->adicity);
      for (x=0; x<f->adicity; x++)
	printf("%d ",f->arg_type[x]->siz);
      for (x=0; x<f->adicity; x++) argument[x] = 0;
      do {
	y = vec_pos(f,argument);
	for (real_value=true,x=0; x<f->adicity; x++)
	  if (argument[x] == f->arg_type[x]->siz) {
	    real_value = false;
	    break;
	  }
	if (real_value)
	  print_cgv_set(domains+y,T->vector[y].csiz);
      }
      while (another_argument(f,argument));
      printf("\n");
    }
}


/*
 * Will be more sophisticated if real CGV output is ever required
 */

void print_cgv_set(set domain, int maxval)
{
  int x;

  printf("{");
  for (x=0; x<maxval; x++)
    if (S_is_member(x,domain))
      printf(" %d",x);
  printf(" } ");
}


/*
 * Basic output of domains for config mode
 */

void print_domains(set possibles[], set impossibles[])
{
  functor f;
  int i;
  int argument[ARGMAX];
  int n;
  boolean toobig;
  FILE *fil = setting.outf;
  ifile ifil = setting.outi;

  xprint(fil,ifil,"DOMAINS\n");
  for (f=FN; f; f = f->nextfunct) 
    if (!f->piped && f->configurable) {
      for (i=0; i<f->adicity; i++) argument[i] = 0;
      if (f->total) n = f->val_type->siz;
      else n = f->val_type->siz+1;
      do {
	toobig = false;
	for (i=0; i<f->adicity; i++)
	  if (f->total && argument[i]==f->arg_type[i]->siz)
	    toobig = true;
	if (!toobig) {
	  xprintfs(fil,ifil,"%s", f->fname);
	  for (i=0; i<f->adicity; i++) {
	    if (i) xprint(fil,ifil,","); else xprint(fil,ifil,"(");
	    printvalue(fil,ifil,f->arg_type[i],argument[i],0,'-');
	  }
	  if (i) xprint(fil,ifil,")");
	  xprint(fil,ifil," Y = {");
	  for (i=0; i<n; i++)
	    if (possibles && S_is_member(i,possibles[vec_pos(f,argument)])) {
	      xprint(fil,ifil," ");
	      printvalue(fil,ifil,f->val_type,i,0,'-');
	    }
	  xprint(fil,ifil," }  N = {");
	  for (i=0; i<n; i++)
	    if (!possibles || S_is_member(i,impossibles[vec_pos(f,argument)])) {
	      xprint(fil,ifil," ");
	      printvalue(fil,ifil,f->val_type,i,0,'-');
	    }
	  xprint(fil,ifil," }\n");
	}
      }
      while (another_argument(f,argument));
    }
  xprint(fil,ifil,"END DOMAINS");
}



void print_set(functor f)
// Debugging function only - protected by DEBUG
{
  int i, j;
  int argument[ARGMAX];

  if (f->adicity > 2) {
    printf("\nAdicity of %s (fcc %d) greater than 2\n",f->fname,f->fcc);
    return;
  }
  printf("\nIndex for %s",f->fname);
  if (!f->adicity) {
    printf(" %d\n", vec_pos(f,argument));
    return;
  }
  puts("");
  if (f->adicity == 1) {
    for (i = 0; i < f->arg_type[0]->siz; i++) {
      argument[0] = i;
      printf("%4d",vec_pos(f,argument));
    }
    argument[0] = i;
    printf("%5d\n",vec_pos(f,argument));
    return;
  }
  for (i = 0; i < f->arg_type[0]->siz; i++) {
    argument[0] = i;
    for (j = 0; j < f->arg_type[1]->siz; j++) {
      argument[1] = j;
      printf("%4d",vec_pos(f,argument));
    }
    argument[1] = j;
    printf("%5d",vec_pos(f,argument));
    puts("");
  }
  puts("");
  argument[0] = i;
  for (j = 0; j < f->arg_type[1]->siz; j++) {
    argument[1] = j;
    printf("%4d",vec_pos(f,argument));
  }
  argument[1] = j;
  printf("%5d",vec_pos(f,argument));
  puts("");
}


void report_literal(intpair lp)
// Was called from local search file. No longer used.
{
  int args[ARGMAX];
  functor f;
  int i;

  f = un_vec_pos(lp->v,args);
  if (f->val_type == Bool) {
    if (lp->w) {
      if (!lp->positive) printf("not ");
      printf("%s",f->fname); if (f->adicity) printf("(");
      for (i=0; i<f->adicity; i++) {
	if (i) printf(",%d",args[i]);
	else printf("%d",args[i]);
      }
      if (f->adicity) printf(")"); printf("\n");
    }
    else {
      if (lp->positive) printf("not ");
      printf("%s",f->fname); if (f->adicity) printf("(");
      for (i=0; i<f->adicity; i++) {
	if (i) printf(",%d",args[i]);
	else printf("%d",args[i]);
      }
      if (f->adicity) printf(")"); printf("\n");
    }
  }
  else {
    if (!lp->positive) printf("not ");
    printf("%s",f->fname); if (f->adicity) printf("(");
    for (i=0; i<f->adicity; i++) {
      if (i) printf(",%d",args[i]);
      else printf("%d",args[i]);
    }
    if (f->adicity) printf(")"); printf(" = %d\n", lp->w);
  }
}


void print_val_of_cell(FILE *fil, ifile ifil, int x, int v)
// Print (by name) value v of the function within which x falls
{
  functor f;
  int farg[ARGMAX];    // Dummy parameter for un_vec_pos

  f = un_vec_pos(x,farg);
  printvalue(fil,ifil,f->val_type,v,0,'-');
}


void print_un_vec_pos(FILE *fil, ifile ifil, int x)
{
  functor f;
  int farg[ARGMAX];
  int i;

  f = un_vec_pos(x,farg);
  if (!f) xprint(fil,ifil,"No function\n");
  else {
    xprintfs(fil,ifil, "%s", f->fname);
    for (i=0; i<f->adicity; i++) {
      if (i) xprint(fil,ifil,","); else xprint(fil,ifil,"(");
      printvalue(fil,ifil,f->arg_type[i],farg[i],0,'-');
    }
    if (i) xprint(fil,ifil,")");
  }
  if (fil) fflush(fil);
}



void print_best_model(int info[], trs T)
{
  vect_into_model(info);
  set_last_model(info);
  updates[current_theory]++;
  workNow = T->job->reps[INVALUES].i_report;
  if (setting.pm.verbosity[pmodels] && !setting.configmode) print_model();
}

#endif


/*
************************************************************************
*   Now the generic functions (not FINDER-specific)
************************************************************************
*/



/*
 * Go through the list of stats collected, and print the numbers etc
 * to standard output.
 */

void dump_report(trin tr)
// Print the end-of-run stats. The files are part of the trin.
// If printing numbers only, just the SOLUTIONS case will happen
{
  int i;
  logical t;

  for (i=0; i<MAX_REPORT && (t = tr->reps[i].report_type); i++) {
    if (i == SOLUTIONS) {
      if (setting.pm.number_only)
	xprintfd(tr->outf,tr->outi, "%d\n", tr->reps[i].i_report);
      else
	xprintfsd(tr->outf,tr->outi,"%s %d\n", tr->reps[i].s_report, tr->reps[i].i_report);
    }
    else if (!(setting.pm.number_only) && (t & INT_REPORT) && tr->reps[i].i_report)
      xprintfsd(tr->outf,tr->outi,"%s %d\n", tr->reps[i].s_report, tr->reps[i].i_report);
    if (!(setting.pm.number_only) && (t & FLOAT_REPORT) && tr->reps[i].f_report)
      xprintfsf(tr->outf,tr->outi,"%s %3.2f\n",tr->reps[i].s_report,tr->reps[i].f_report);
  }
  for (i=0; i<MAX_REPORT && (t = tr->reps[i].report_type); i++)
    if (!(setting.pm.number_only) && (t & DO_REPORT))
      (*tr->reps[i].do_report)(tr);
}


/*
 * Print just the vector of numbers, for runtime profiling
 */

void dump_trace(trin tr, boolean final)
// Happens if tr->dump_rate is set. Only for stdout output channel.
{
  int i;
  logical t;
  int x = 0;

  tr->tracepoints++;
  if (final || !(tr->tracepoints % tr->dump_rate)) {
    if (final) printf("-1\n");
    for (i=0; i<MAX_REPORT; i++) {
      t = tr->reps[i].report_type;
      if (t & INT_REPORT) {
	if (final) printf("%d. %s (int)\n", ++x, tr->reps[i].s_report);
	else printf(" %d", tr->reps[i].i_report);
      }
      if (t & FLOAT_REPORT) {
	if (final) printf("%d. %s (float)\n", ++x, tr->reps[i].s_report);
	else printf(" %0.2f", tr->reps[i].f_report);
      }
    }
    printf("\n");
  }
}


/*
 * The next two functions print the constraint graph in the right
 * format to be used by CGV.
 */

void print_constraint_graph(trs T)
{
  int x, y;
  reflist r;
  FILE *fil = T->job->outf;
  ifile ifil = T->job->outi;

  if (!(T->job->profile)) return;

  xprint(fil,ifil,"g12vis start\n");
  /* Report the variables */
  for (x=0; x<T->job->vlength; x++)
    if (T->vector[x].poss > 1) {
      xprintfd(fil,ifil,"add var v%d => ", x);
      print_un_vec_pos(fil,ifil,x);
      xprint(fil,ifil,"\n");
    }

  /* Report the constraints */
  for (x=0; x<T->job->vlength; x++) {
    for (y=0; y<T->vector[x].csiz; y++) {
      for (r = T->vector[x].cvals[y].cvrefs; r; r = r->nextref)
        check_new_ref(x, r->rf, T);
      for (r = T->vector[x].cvals[y].posrefs; r; r = r->nextref)
        check_new_ref(x, r->rf, T);
    }
    for (r = T->vector[x].inject; r; r = r->nextref)
      check_new_ref(x, r->rf, T);
  }
  xprint(fil,ifil,"g12vis end\n");
}


/*
 * Subroutine of the above: if the ref is new, add a constraint to the
 * constraint graph.
 */

void check_new_ref(int x, ref rf, trs T)
{
  intpair il;
  int count = 0;
  FILE *fil = T->job->outf;
  ifile ifil = T->job->outi;

  if (!rf) return;
  for (il=rf->cellvals; il; il=il->p) {
    if (il->v < x)
      return;
    if (T->vector[il->v].cvals[il->w].blocks && !(il->p))
      return;
    if (T->vector[il->v].poss == 1) {
      if (!(T->vector[il->v].cvals[il->w].blocks) && il->p)
        return;
    }
    else
      count++;
  }

  if (count>1) {
    xprintfld(fil,ifil,"add con c%ld [", (long int)rf);
    for (il=rf->cellvals; il; il=il->p)
      if (T->vector[il->v].poss > 1)
        xprintfd(fil,ifil," v%d", il->v);
    xprint(fil,ifil," ]\n");
  }
}


void report_cell_val(intpair lp)
// Not currently called from amywhere. Pertains to local search.
{
  printf(" %d",lp->v);
  if (lp->positive) printf("="); else printf("#");
  printf("%d",lp->w);
  if (lp->p) printf(",");
}


void report_constraint(ref thisref)
// All calls (from fvntr.c) protected by DEBUG
{
  intpair lp;

  printf("%ld\n",(long int)thisref);
  for (lp=thisref->cellvals; lp; lp = lp->p) {
#ifdef IN_FINDER
    report_literal(lp);
#else
    report_cell_val(lp);
#endif
  }
  if (thisref->softweight) printf("soft "); else printf("hard ");
  if (thisref->notop) printf("reduced ");
  if (thisref->violated) printf("violated"); else printf("not violated");
  printf("\n");
  fflush(stdout);
}

void report_constraint_list(reflist theserefs)
// Not currently called from amywhere. Prints a list of low-level constraints.
{
  reflist r;

  for (r=theserefs; r; r=r->nextref) {
    report_constraint(r->rf);
    printf("\n");
  }
  fflush(stdout);
}



void report_branches(trin tr)
// Part of printing the final stats
{
  int i, tsz = 0;

  if (tr->branches) {
    for (tsz=tr->maxval; tsz && !tr->branches[tsz]; tsz--) ;
    if (tsz) {
      xprintfd(tr->outf,tr->outi,"Branching: [%d]",tr->branches[1]);
      for (i=2; i <= tsz; i++)
	xprintfd(tr->outf,tr->outi," %d",tr->branches[i]);
      xprint(tr->outf,tr->outi,"\n");
    }
  }
}


void stopped_by(trin tr)
// Printed along with the final stats
{
  if (tr->done)
    xprintfs(tr->outf,tr->outi,"Stopped by %s\n",tr->reps[STOP_CODE].s_report);
  else
    xprint(tr->outf,tr->outi,"Stopped by exhaustion\n");
}


void dump_space(trs T)
// For each functor, report how many cells remain with each number of
// possible values
{
  functor f;
  int arguments[ARGMAX];
  int count[31];
  int m,n;
  int x;

  for (f = FN; f; f = f->nextfunct) {
    for (x=0; x<31; x++) count[x] = 0;
    if (f->total) n = f->val_type->siz;
    else n = f->val_type->siz+1;
    if (n > 30) n = 31;
    xprintfs(T->job->outf,T->job->outi,"%s ",f->fname);
    for (x=0; x<ARGMAX; x++) arguments[x] = 0;
    do {
      x = T->vector[vec_pos(f,arguments)].poss;
      if (x <= 30) count[x]++;
      else count[31]++;
    }
    while (another_argument(f,arguments));
    for (m=n+1; m && count[m-1]==0; m--) ;
    for (x=0; x<m; x++) {
      xprintfd(T->job->outf,T->job->outi," %d:", x);
      xprintfd(T->job->outf,T->job->outi,"%d", count[x]);
    }
    xprint(T->job->outf,T->job->outi," ]\n");
  }
  xprint(T->job->outf,T->job->outi," ]\n");
}
