#define IFILES

/* Number of bytes allocated each time an ifile is extended */
#ifndef ILEN
#define ILEN 1024
#endif

/* Maximum ifile size - can be overridden in the makefile */
#ifndef ILENLIMIT
#define ILENLIMIT (1000*ILEN)
#endif

/* Integer for use in iscanf */
#ifdef IPRINTLIB
int i_read_length;
#else
extern int i_read_length;
#endif
#ifndef IEND
#define IEND &i_read_length
#endif

/* Trivial macro to zero the pointer after closing an ifile */
#ifndef ICLOSE
#define ICLOSE
#define iclose(i) _iclose(i); i=0
#endif

#ifndef IFILESTRUCT
#define IFILESTRUCT
/*
The ifile structure. It behaves essentially like a file, but is held
 in memory. Formatted text can be appended to it just as it can to any
 output stream, and for an ifile i, the string (i->s)+(i->r) may be
 treated as an input stream. In reading formatted input by means of
 iscanf, the dummy parameter IEND must be added after the list of
 pointers to which the data is assigned, and the corresponding %n to
 the end of the format string. Format strings etc are otherwise
 exactly as in the scanf and printf family.
 */

typedef struct {
  char *s;        // Actual contents
  int s_len;      // Maximum string length
  int r;          // Read point
} IFILE, *ifile;

/* Also for convenience, [lists of] pointers to indexed ifiles */

typedef struct ifile_ptr_t {
  int id;
  ifile f;
  struct ifile_ptr_t *last;
  struct ifile_ptr_t *next;
} IFILE_PTR, *ifile_ptr;

typedef struct {
  int len;
  ifile_ptr start;
  ifile_ptr end;
} IFILE_LIST, *ifile_list;

#define isopen(x,l) (itof(x,l) != 0)

#endif


/* Function prototypes */
ifile iopen(void);
// Set up a new ifile for reading and writing. Return a pointer to it.

void ireread(ifile i);
// Set the input point to start of file

ifile ifile_init(ifile i);
// If i already points at an ifile, zero it. If not, allocate one and zero it.
// Return a pointer to the newly initialised ifile (i.e. return i).

ifile_list init_ifile_list(ifile_list L);
// If L already points at an ifile list, zero it. If not, allocate one and zero it.
// Return a pointer to the newly initialised ifile list (i.e. return L).

ifile remove_ifile(int i, ifile_list L);
// What it says: remove ifile with index i from L but don't close it.
// Return a pointer to the removed ifile

int add_ifile(ifile f, ifile_list L);
// Ensure f is indexed in list L, and return its index

ifile itoif(int x, ifile_list L);
// Point to ifile with index x

int iftoi(ifile f, ifile_list L);
// Return index of ifile f, or 0 if not in the list

ifile new_ifile(ifile_list L);
// Initialise a fresh ifile, add it to L, return a pointer to it

void ifile_extend(ifile i);
// Allocate a new chunk of memory and append it to the ifile string

int iputc(int c, ifile i);
// The analogue of fputc. Returns the character written

int iputs(char *s, ifile i);
// The analogue of fputs. Print s plus a terminating newline character to i
// Return number of characters printed, or EOF on error

int iprintf(ifile i, const char* format, ... );
// The analogue of fprintf. Return the number of characters added

char igetc(ifile i);
// The analogue of fgetc. Returns a char rather than an unsigned char cast to int.

char* igets(char *str, int size, ifile i);
// The analogue of fgets. Returns a pointer to the string, or NULL on failure.

int iscanf(ifile i, const char* format, ... );
// The analogue of fscanf. It is assumed that IEND is the last argument in the
// ellipsis and that the last conversion string in the format is %n.
// Return the number of input items assigned

void _iclose(ifile i);
// Delete an ifile and free the memory. It cannot be re-opened after this.
// Use iclose instead wherever possible, to prevent junk references.

void dumpitof(ifile i, FILE *f);
// Similar to cat: transfer contents of i to f. Don't open or close files.
// I/O is character-by-character. This is not prohibitively slow in C.
