/* Handle virtual files (in memory) */

#define IPRINTLIB

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#ifndef IFILES
#include "ifile.h"
#endif



ifile iopen(void)
// Set up a new ifile for reading and writing. Return a pointer to it.
{
  return ifile_init(0);
}

void ireread(ifile i) { if (i) i->r = 0; }
// Set the input point to start of file

ifile ifile_init(ifile i)
// If i already points at an ifile, zero it. If not, allocate one and zero it.
// Return a pointer to the newly initialised ifile (i.e. return i).
{
  iclose(i);
  i = (ifile) malloc(sizeof(IFILE));
  i->s = 0;
  i->s_len = 0;
  i->r = 0;
  return i;
}

/* Basic memory management for lists of ifiles */

ifile_list init_ifile_list(ifile_list L)
// If L already points at an ifile list, zero it. If not, allocate one and zero it.
// Return a pointer to the newly initialised ifile list (i.e. return L).
{
  if (!L)
    L = (ifile_list) malloc(sizeof(IFILE_LIST));
  else
    while(L->len > 0)
      remove_ifile(L->start->id,L);
  L->len = 0;
  L->start = L->end = 0;
  return L;
}

ifile remove_ifile(int i, ifile_list L)
// What it says: remove ifile with index i from L but don't close it.
// Return a pointer to the removed ifile
{
  ifile_ptr ptr;
  ifile rc;

  for (ptr=L->start; ptr; ptr=ptr->next)
    if (ptr->id == i) {
      if (ptr->last)
	ptr->last->next = ptr->next;
      else
	L->start = ptr->next;
      if (ptr->next)
	ptr->next->last = ptr->last;
      else
	L->end = ptr->last;
      L->len--;
      rc = ptr->f;
      ptr->f = 0;   // Not needed: initialise to prevent any double
      ptr->id = 0;  // free in case of a bug (not that bugs exist...)
      ptr->next = ptr->last = 0;
      free(ptr);
      return rc;
    }
  return 0;
}

int add_ifile(ifile f, ifile_list L)
// Ensure f is indexed in list L, and return its index
{
  ifile_ptr ptr;
  int maxid = 0;

  for (ptr=L->start; ptr; ptr=ptr->next)
    if (ptr->f == f)
      return ptr->id;
    else if (ptr->id > maxid)
      maxid = ptr->id;
  ptr = (ifile_ptr) malloc(sizeof(IFILE_PTR));
  ptr->next = 0;
  ptr->last = L->end;
  ptr->id = maxid+1;
  ptr->f = f;
  if (L->len == 0) {
    L->start = L->end = ptr;
  }
  else
    L->end->next = ptr;
  L->len++;
  return ptr->id;
}

ifile itoif(int x, ifile_list L)
// Point to ifile with index x
{
  ifile_ptr ptr;

  for (ptr=L->start; ptr; ptr=ptr->next)
    if (ptr->id == x)
      return ptr->f;
  return 0;
}

int iftoi(ifile f, ifile_list L)
// Return index of ifile f, or 0 if not in the list
{
  ifile_ptr ptr;

  for (ptr=L->start; ptr; ptr=ptr->next)
    if (ptr->f == f)
      return ptr->id;
  return 0;
}

ifile new_ifile(ifile_list L)
// Initialise a fresh ifile, add it to L, return a pointer to it
{
  return itoif(add_ifile(iopen(),L),L);
}

void ifile_extend(ifile i)
// Allocate a new chunk of memory and append it to the ifile string
{
  if (!i) return;
  if (!(i->s)) {
    i->s = (char*) malloc(ILEN*sizeof(char));    // Size of char should be 1, but anyway...
    *(i->s) = 0;
    i->s_len = ILEN;
  }
  else {
    i->s_len = strlen(i->s)+ILEN;
    if (i->s_len > ILENLIMIT) {
      fprintf(stderr,"Error: length limit for internal file exceeded\n");
      exit(1);
    }
    i->s = (char*) realloc(i->s,i->s_len*sizeof(char));
  }
}


int iputc(int c, ifile i)
// The analogue of fputc. Returns the character written
{
  int input_point;

  input_point = strlen(i->s);
  if (input_point >= i->s_len)
    ifile_extend(i);
  i->s[input_point] = (unsigned char)c;
  i->s[input_point+1] = 0;
  return c;
}

int iputs(char *s, ifile i)
// The analogue of fputs. Print s plus a terminating newline character to i
// Return number of characters printed, or EOF on error
{
  if (!i) return -1;
  if (!s) return iputs("(null)",i);
  while (!(i->s) || strlen(i->s) + strlen(s) >= i->s_len-1)
    ifile_extend(i);
  strcat(i->s,s);
  iputc('\n',i);
  return strlen(s)+1;
}


int iprintf(ifile i, const char* format, ... )
// The analogue of fprintf. Return the number of characters added
{
  va_list arglist;
  char *newstring = 0;
  int rc;
  
  if (!i || !format || !*format) return 0;
  va_start( arglist, format );
  rc = vasprintf(&newstring,format,arglist);
  va_end( arglist );
  if (newstring) {
    while (!(i->s) || strlen(i->s) + strlen(newstring) >= i->s_len)
      ifile_extend(i);
    strcat(i->s,newstring);
    free(newstring);
  }
  return rc;
}

char igetc(ifile i)
// The analogue of fgetc. Returns a char rather than an unsigned char cast to int.
{
  if (!i || !(i->s) || i->r >= strlen(i->s)) return 0;
  i->r++;
  return i->s[i->r-1];
}

char* igets(char *str, int size, ifile i)
// The analogue of fgets. Returns a pointer to the string, or NULL on failure.
{
  int x;
  int sls;

  if (!i || !i->s) return 0;      // Bad input: return NULL
  sls = strlen(i->s);
  if (i->r >= sls) return 0;      // EOF reached: return NULL
  for (x=0; x < size && i->r < sls && i->s[i->r] != '\n'; x++,i->r++)
    str[x] = i->s[i->r];
  if (x < size && i->s[i->r] == '\n')
    str[x++] = i->s[(i->r)++];
  str[x] = 0;
  return str;
}

int iscanf(ifile i, const char* format, ... )
// The analogue of fscanf. Difficulty is to know where to put the read
// point.  It is assumed that IEND is the last argument in the
// ellipsis and that the last conversion string in the format is %n.
// 
// If there were a reliable way to extend a va_list, this wouldn't be
// necessary, as it could be added after the va_start.
{
  va_list arglist;
  char *buf = (i->s)+(i->r);
  int rc;

  if (!i || !format || !*format) return 0;
  if (strcmp(format+(strlen(format)-2),"%n")) {
    fprintf(stderr,"Error reading from internal file: possibly missing \"%cn\"\n",'%');
    exit(1);
  }
  i_read_length = 0;
  va_start( arglist, format );
  rc = vsscanf( buf, format, arglist );
  if (rc > i_read_length) {
    fprintf(stderr,"Error reading from internal file: possible missing IEND\n");
    exit(1);
  }
  i->r += i_read_length;
  va_end( arglist );
  return rc;
}


void _iclose(ifile i)
// Delete an ifile and free the memory. It cannot be re-opened after this.
// Use iclose instead wherever possible, to prevent junk references.
{
  if (i) {
    if (i->s) {
      free(i->s);
      i->s = 0;
    }
    free(i);
  }
}



void dumpitof(ifile i, FILE *f)
// Similar to cat: transfer contents of i to f. Don't open or close files.
// I/O is character-by-character. This is not prohibitively slow in C.
{
  char c;

  if (!i | !f) return;
  ireread(i);
  do {
    c = igetc(i);
    fputc(c,f);
  }
  while (c);
}
